#A* -------------------------------------------------------------------
#B* This file contains source code for the PyMOL computer program
#C* copyright 1998-2000 by Warren Lyford Delano of DeLano Scientific. 
#D* -------------------------------------------------------------------
#E* It is unlawful to modify or remove this copyright notice.
#F* -------------------------------------------------------------------
#G* Please see the accompanying LICENSE file for further information. 
#H* -------------------------------------------------------------------
#I* Additional authors of this source file include:
#-* 
#-* 
#-*
#Z* -------------------------------------------------------------------

import cmd
import math
import string
import pymol
import os
from pymol import movie
# legacy mappings, remove in PyMOL 2.0

mload = movie.load
mrock = movie.rock
mroll = movie.roll

def cbag(selection="(all)"):
   s = str(selection)
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("carbon","(elem C and "+s+")")

def cbac(selection="(all)"):
   s = str(selection)
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("cyan","(elem C and "+s+")")

def cbay(selection="(all)"):
   s = str(selection)   
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("yellow","(elem C and "+s+")")

def cbas(selection="(all)"):
   s = str(selection)   
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("salmon","(elem C and "+s+")")

def cbap(selection="(all)"):
   s = str(selection)   
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("purple","(elem C and "+s+")")

def cbak(selection="(all)"):
   s = str(selection)   
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("pink","(elem C and "+s+")")

def cbaw(selection="(all)"):
   s = str(selection)   
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("hydrogen","(elem C and "+s+")")

def cbab(selection="(all)"):
   s = str(selection)   
   cmd.color("magenta","("+s+")")
   cmd.color("oxygen","(elem O and "+s+")")
   cmd.color("nitrogen","(elem N and "+s+")")
   cmd.color("sulfur","(elem S and "+s+")")
   cmd.color("hydrogen","(elem H and "+s+")")
   cmd.color("slate","(elem C and "+s+")")

def performance(mode):
   mode = int(mode)
   if mode==0: # maximum quality
      cmd.set('line_smooth',1)
      cmd.set('depth_cue',1)         
      cmd.set('specular',1)
      cmd.set('surface_quality',1)
      cmd.set('stick_quality',15)
      cmd.set('sphere_quality',2)
      cmd.set('cartoon_sampling',14)
      cmd.set('ribbon_sampling',10)
      cmd.do("rebuild")
   elif mode==33:
      cmd.set('line_smooth',1)         
      cmd.set('depth_cue',1)         
      cmd.set('specular',1)
      cmd.set('surface_quality',0)
      cmd.set('stick_quality',8)
      cmd.set('sphere_quality',1)
      cmd.set('cartoon_sampling',7)
      cmd.do("rebuild")
   elif mode==66: # good perfomance
      cmd.set('line_smooth',0)
      cmd.set('depth_cue',0)         
      cmd.set('specular',1)
      cmd.set('surface_quality',0)
      cmd.set('stick_quality',8)
      cmd.set('sphere_quality',1)
      cmd.set('cartoon_sampling',6)
      cmd.do("rebuild")         
   else: # maximum performance
      cmd.set('line_smooth',0)
      cmd.set('depth_cue',0)
      cmd.set('specular',0)
      cmd.set('surface_quality',-1) # new
      cmd.set('stick_quality',5)
      cmd.set('sphere_quality',0)
      cmd.set('cartoon_sampling',3)
      cmd.do("rebuild")         

def hide_sele():
   arg = cmd.get_names("selections")
   for a in arg:
      cmd.disable(a)

# FUBAR
#def hbond(a,b,cutoff=3.3,name='hbond'):
#   cmd.dist(name,"((%s) and ((%s) around %4.2f) and elem N,O)"%(a,b,cutoff),
#            "((%s) and ((%s) around %4.2f) and elem N,O)"%(b,a,cutoff),
#            cutoff)

def cbc(selection='(all)',first_color=7): # NOT THREAD SAFE
   '''
   Color all chains a different color
   '''
   pymol.stored.chain = {}
   cmd.iterate("(%s)"%selection,"stored.chain[chain]=1")
   c = first_color
   for a in pymol.stored.chain.keys():
      if len(a):
         print ("%d,(chain %s)"%(c,a))
         cmd.color("%d"%c,"(chain %s)"%a)
         c = c + 1
      else:
         print ("%d,(chain '')"%(c))
         cmd.color("%d"%c,"(chain '')")
         c = c + 1
         
color_chains = cbc

def sum_charge(*arg): # NOT THREAD SAFE
   result = None
   try:
      obj = "all"
      if len(arg):
         obj = arg

      pymol.stored._sum_charge = 0.0
      cmd.iterate("(%s)"%obj,
                  "stored._sum_charge=stored._sum_charge+partial_charge")
      result = pymol.stored._sum_charge
      print " sum_charge: %6.4f"%result
   except:
      print " sum_charge: an error occurred."
   return result

def ray_shadows(mode):
   if mode=='light': # maximum quality
      cmd.set('ambient',0.3)
      cmd.set('spec_reflect',0.4)
      cmd.set('spec_power',40)
      cmd.set('reflect',1.2)
      cmd.set('reflect_power',1.0)         
      cmd.set('direct',0.35)
      cmd.set('power',3.0)
      cmd.set('gamma',1.2)
   elif mode=='matte':
      cmd.set('spec_reflect',0.4)
      cmd.set('spec_power',20)
      cmd.set('ambient',0.12)
      cmd.set('reflect',0.9)
      cmd.set('reflect_power',1.0)
      cmd.set('power',1.5) 
      cmd.set('direct',0.25)
      cmd.set('gamma',1.30)
   elif mode=='medium':
      cmd.set('power',1.0) # 0.7
      cmd.set('reflect_power',1.0)
      cmd.set('spec_reflect',0.5)
      cmd.set('spec_power',60) # was 50
      cmd.set('ambient',0.12)
      cmd.set('reflect',0.9) 
      cmd.set('direct',0.25) 
      cmd.set('gamma',1.3) 
   elif mode=='heavy':
      cmd.set('reflect_power',1.0)
      cmd.set('spec_reflect',0.60) # was 0.53
      cmd.set('spec_power',90) # was 60
      cmd.set('ambient',0.03) # was 0.08
      cmd.set('reflect',0.75) # was 0.70
      cmd.set('direct',0.06)
      cmd.set('power',0.3)
      cmd.set('gamma',1.43) # was 1.5
   elif mode=='black': # best for light backgrounds
      cmd.set('power',0.3)
      cmd.set('spec_reflect',1.1)
      cmd.set('spec_power',90) 
      cmd.set('ambient',0.00)
      cmd.set('reflect',0.8)
      cmd.set('direct',0.00)
      cmd.set('reflect_power',1.7)
      cmd.set('gamma',1.55) 

def ff_copy(src,dst): # NOT THREAD SAFE
   pymol._rcopy = pymol.Scratch_Storage()
   pymol._rcopy.pc={}
   pymol._rcopy.tt={}
   cmd.iterate("(%s)"%src,"_rcopy.pc[name]=partial_charge")
   cmd.alter("(%s)"%dst,"partial_charge=_rcopy.pc[name]")
   cmd.iterate("(%s)"%src,"_rcopy.tt[name]=text_type")
   cmd.alter("(%s)"%dst,"text_type=_rcopy.tt[name]")
   del pymol._rcopy
   
def b2vdw(*arg):
   if not len(arg):
      sele = 'all'
   else:
      sele = arg[0]
   # use B values to create RMS VDW spheres
   # rms = sqrt(b/(8*(PI^2)))
   cmd.alter("(%s)"%sele,"vdw=math.sqrt(b/78.9568352087)")
   
def phipsi(selection="(pk1)"): # NOT THREAD SAFE
   n_sele =   "((byres (%s)) & name n)"%selection
   c_sele =   "((byres (%s)) & name c)"%selection
   ca_sele =  "((byres (%s)) & name ca)"%selection
   cm_sele = "((neighbor (%s)) and not (byres (%s)))"%(n_sele,n_sele)
   np_sele = "((neighbor (%s)) and not (byres (%s)))"%(c_sele,c_sele)
   cmd.feedback("push")
   cmd.feedback("disable","selector","everythin")
   cm_cnt = cmd.select("_pp_cm",cm_sele)
   n_cnt = cmd.select("_pp_n",n_sele)
   c_cnt = cmd.select("_pp_c",c_sele)
   ca_cnt = cmd.select("_pp_ca",ca_sele)
   np_cnt = cmd.select("_pp_np",np_sele)
   if(cm_cnt and n_cnt and ca_cnt and c_cnt):
      phi = cmd.get_dihedral("_pp_c","_pp_ca","_pp_n","_pp_cm")
   else:
      phi = None
   if(n_cnt and ca_cnt and c_cnt and np_cnt):
      psi = cmd.get_dihedral("_pp_np","_pp_c","_pp_ca","_pp_n")
   else:
      psi = None
   cmd.feedback("pop")
   cmd.delete("_pp_cm")
   cmd.delete("_pp_n")
   cmd.delete("_pp_c")
   cmd.delete("_pp_ca")
   cmd.delete("_pp_np")
   return (phi,psi)

def rainbow(selection="(name ca and alt '',A)",reverse=0): # NOT THREAD SAFE

   cmd.feedback("push")
   cmd.feedback("disable","executive","actions")

   # your basic rainbow...
   
   list = [
      (0,0,255),      
      (0,0,255),
      (0,128,255),
      (0,255,255),
      (0,255,128),           
      (0,255,0),
      (128,255,0),
      (255,255,0),
      (255,128,0),
      (255,0,0),
      (255,0,0)      
      ]
   if reverse:
      list.reverse()
   #
   last = list.pop(0)
   cmd.set_color("_000",[last[0]/255.0,last[1]/255.0,last[2]/255.0])
   c = 1
   for a in list:
      for b in range(1,21):
         b0 = b/20.0
         b1 = 1.0-b0
         cname = "_%03d"%c
         r = last[0]*b1+a[0]*b0
         g = last[1]*b1+a[1]*b0
         b = last[2]*b1+a[2]*b0
         cmd.set_color(cname,[r/255.0,g/255.0,b/255.0])
         c = c + 1
      last = a

   cas = cmd.index("((byres ("+selection+")) and name ca and not het)")
   l = len(cas)
   if not len(cas):
      return
   c = 0
   for a in cas:
      col = int((200*c)/l)
      cmd.color("_%03d"%col,"((%s) and (byres %s`%d))"%(selection,a[0],a[1]))
      c = c + 1

   cmd.feedback("pop")
   
def ss(selection="(name ca and alt '',A)",state=1): # NOT THREAD SAFE

   print ' util.ss: WARNING: This is not a "correct" secondary structure'
   print ' util.ss: assignment algorithm!  Please use only as a last resort.'
   
   cmd.feedback("push")
   cmd.feedback("disable","executive","actions")
   
   ss_pref = "_sss"
   sss1 = ss_pref+"1"
   cnt = cmd.select(sss1,"((byres ("+selection+")) and name ca and not het)")
   print " util.ss: initiating secondary structure assignment on %d residues."%cnt
   cas = cmd.index(sss1)
   if not len(cas):
      return
   # set cartoon mode to auto over the selection
   
   cmd.cartoon("auto",sss1)

   print " util.ss: extracting sequence and relationships..."

   # get CA list
   
   res_list = []
   pymol._ss = pymol.Scratch_Storage()
   pymol._ss.res_list = res_list
   cmd.iterate(sss1,'_ss.res_list.append((model,index))')

   # generate atom-to-residue conversion dictionaries

   ca_dict = {}
   n_dict = {}
   o_dict = {}
   scr_dict = {} # scr = segment,chain,resi 
   pymol._ss.n_dict = n_dict
   pymol._ss.o_dict = o_dict
   pymol._ss.scr_dict = scr_dict
   pymol._ss.ca_dict = ca_dict
   cmd.iterate(sss1,
               '_ss.scr_dict[(model,index)]=(segi,chain,resi)') # CA's
   cmd.iterate("((byres "+sss1+") and n;n)"
               ,'_ss.scr_dict[(model,index)]=(segi,chain,resi)') # N's
   cmd.iterate("((byres "+sss1+") and n;o)",
               '_ss.scr_dict[(model,index)]=(segi,chain,resi)') # O's
   cmd.iterate(sss1,
               '_ss.ca_dict[(segi,chain,resi)] = (model,index)')
   cmd.iterate("((byres "+sss1+") and n;n)",
               '_ss.n_dict[(segi,chain,resi)] = (model,index)')
   cmd.iterate("((byres "+sss1+") and n;o)",
               '_ss.o_dict[(segi,chain,resi)] = (model,index)')

   scr_dict[None]=None
   o_dict[None]=None
   n_dict[None]=None
   ca_dict[None]=None
   
   # create special version of cas with gaps

   gap = [None,None,None,None]  
   # gap large enough to distinguish i+4 interations from gaps
   last = None
   for a in res_list:
      if last!=None:
         if(cmd.count_atoms(
            "((neighbor(neighbor(neighbor (%s`%d)))) and (%s`%d))"%
            (last[0],last[1],a[0],a[1]),quiet=1)==0):
            gap.extend([None,None,None,None])
      gap.append(a)
      last = a
   gap.extend([None,None,None,None])

   print " util.ss: analyzing phi/psi angles (slow)..."

   # generate reverse-lookup for gap indices

   ss = {}

   c = 0
   gap_idx = {}
   for a in gap:
      gap_idx[a] = c
      c = c + 1

   # secondary structure database...
   
   ss = {}
   ss[None]=None
   
   # make decisions based on phi/psi

   for a in cas:
      ss[a] = 'L' # default
   phipsi = cmd.get_phipsi(sss1,state)
   for a in phipsi.keys():
      (phi,psi) = phipsi[a]
#      print scr_dict[a],(phi,psi)
      if (phi!=None) and (psi!=None):
         if ((phi<-45) and (phi>-160) and
             (psi<-170) or (psi>10)): # beta?
            ss[a] = 's'
         elif ((phi<-45) and (phi>-160) and
               (psi>-80) and (psi<-25)): # helix?
            ss[a] = 'H'
            
   print " util.ss: finding hydrogen bonds..."
   
   # find all pairwise hydrogen bonds and make note of them in dict

   hb = cmd.find_pairs("((byres "+sss1+") and n;n)",
                       "((byres "+sss1+") and n;o)",mode=1,
                       cutoff=3.7,angle=55,
                       state1=state,state2=state)
   
   hb_dict = {}  # [((N-atom) (O-atom))] = 1
   n_hb_dict = {} # [(N-atom)] = [(O-atom),...]
   o_hb_dict = {} # [(O-atom)] = [(N-atom),...]
   for a in hb:
#      cmd.dist("(%s`%d)"%a[0],"(%s`%d)"%a[1])
      hb_dict[a] = 1
      n = a[0]
      o = a[1]
      if not n_hb_dict.has_key(n): n_hb_dict[n]=[]
      if not o_hb_dict.has_key(o): o_hb_dict[o]=[]
      n_hb_dict[n].append(o)
      o_hb_dict[o].append(n)

   # check to insure that all helical residues have at least an i +/- 4
   # hydrogen bond

   for c in xrange(4,len(gap)-4):
      a = gap[c]
      if ss[a]=='H':
         aN = n_dict[scr_dict[a]]
         aO = o_dict[scr_dict[a]]
         am4O = o_dict[scr_dict[gap[c-4]]]
         ap4N = n_dict[scr_dict[gap[c+4]]]
         if not hb_dict.has_key((aN,am4O)):
            if not hb_dict.has_key((ap4N,aO)):
               ss[a]='L'

   print " util.ss: verifying beta sheets..."
   
   # check to insure that all beta residues have proper interactions

   rep_dict = {}
   repeat = 1
   while repeat:
      repeat = 0
      c = 4
      cc = len(gap)-4
      while c<cc:
         a1 = gap[c]
         if (ss[a1] in ['s','S']) and not rep_dict.has_key(a1):
            rep_dict[a1] = 1
            valid = 0
            scr_a1 = scr_dict[a1]
            # look for antiparallel 2:2 H-bonds (NH-O=C + C=O-HN) 
            n_a1_atom = n_dict[scr_a1]
            o_a1_atom = o_dict[scr_a1]
            if (n_hb_dict.has_key(n_a1_atom) and 
                o_hb_dict.has_key(o_a1_atom)):
               for n_hb_atom in n_hb_dict[n_a1_atom]:
                  for o_hb_atom in o_hb_dict[o_a1_atom]:
                     n_hb_scr = scr_dict[n_hb_atom]
                     o_hb_scr = scr_dict[o_hb_atom]
                     if o_hb_scr == n_hb_scr:
                        b1 = ca_dict[o_hb_scr]
                        if abs(c-gap_idx[b1])>2:
                           ss[b1] = 'S' 
                           ss[a1] = 'S' 
                           valid = 1
            # look for antiparallel offset HB (i,i+2,j,j-2)
            a3 = gap[c+2]
            if (a3!=None):
               scr_a3 = scr_dict[a3]
               o_a1_atom = o_dict[scr_a1]
               n_a3_atom = n_dict[scr_a3]
               if (n_hb_dict.has_key(n_a3_atom) and
                   o_hb_dict.has_key(o_a1_atom)):               
                  for n_hb_atom in n_hb_dict[n_a3_atom]:
                     for o_hb_atom in o_hb_dict[o_a1_atom]:
                        n_hb_scr = scr_dict[n_hb_atom]
                        o_hb_scr = scr_dict[o_hb_atom]
                        b1 = ca_dict[o_hb_scr]
                        if b1!=None:
                           b1_i = gap_idx[b1]
                           if abs(c-b1_i)>2: # no turns!
                              b3 = gap[b1_i-2]
                              if b3!=None:
                                 b3_scr = scr_dict[b3]
                                 if b3_scr == n_hb_scr:
                                    a2 = gap[c+1]
                                    b2 = gap[gap_idx[b1]-1]
                                    ss[b1] = 'S'
                                    ss[b3] = 'S'
                                    ss[a1] = 'S'
                                    ss[a3] = 'S'
                                    if ss[a2]=='L': ss[a2] = 's'
                                    if ss[b2]=='L': ss[b2] = 's'
                                    valid = 1
            # look for antiparallel offset HB (i,i-2,j,j+2)
            a3 = gap[c-2]
            if (a3!=None):
               scr_a3 = scr_dict[a3]
               n_a1_atom = n_dict[scr_a1]
               o_a3_atom = o_dict[scr_a3]
               if (n_hb_dict.has_key(n_a1_atom) and
                   o_hb_dict.has_key(o_a3_atom)):               
                  for n_hb_atom in n_hb_dict[n_a1_atom]:
                     for o_hb_atom in o_hb_dict[o_a3_atom]:
                        n_hb_scr = scr_dict[n_hb_atom]
                        o_hb_scr = scr_dict[o_hb_atom]
                        b1 = ca_dict[o_hb_scr]
                        if b1!=None:
                           b1_i = gap_idx[b1]
                           if abs(c-b1_i)>2: # no turns!
                              b3 = gap[b1_i-2]
                              if b3!=None:
                                 b3_scr = scr_dict[b3]
                                 if b3_scr == n_hb_scr:
                                    a2 = gap[c-1]
                                    b2 = gap[gap_idx[b1]-1]
                                    ss[b1] = 'S'
                                    ss[b3] = 'S'
                                    ss[a1] = 'S'
                                    ss[a3] = 'S'
                                    if ss[a2]=='L': ss[a2] = 's'
                                    if ss[b2]=='L': ss[b2] = 's'
                                    valid = 1
            # look for parallel 1:3 HB (i,j-1,j+1)
            n_a1_atom = n_dict[scr_a1]
            o_a1_atom = o_dict[scr_a1]
            if (n_hb_dict.has_key(n_a1_atom) and
                o_hb_dict.has_key(o_a1_atom)):
               for n_hb_atom in n_hb_dict[n_a1_atom]:
                  for o_hb_atom in o_hb_dict[o_a1_atom]:
                     n_hb_scr = scr_dict[n_hb_atom]
                     o_hb_scr = scr_dict[o_hb_atom]
                     b0 = ca_dict[n_hb_scr]
                     if b0!=None:
                        b2 = gap[gap_idx[b0]+2]
                        if b2!=None:
                           b2_scr = scr_dict[b2]
                           if b2_scr == o_hb_scr:
                              b1 = gap[gap_idx[b0]+1]
                              ss[a1] = 'S' 
                              ss[b0] = 'S'
                              if ss[b1]=='L': ss[b1]='s'
                              ss[b2] = 'S'
                              valid = 1
                              repeat = 1
            if not valid:
               ss[a1] = 'L'
         c = c + 1

   # automatically fill 1 residue gaps in helices and well-defined sheets
   c = 4
   cc = len(gap)-6
   while c<cc:
      a1 = gap[c]
      a3 = gap[c+2]
      ss_a1 = ss[a1]
      ss_a3 = ss[a3]
      if (ss_a1==ss_a3) and (ss_a1 in ['S','H']):
         a2 = gap[c+1]
         ss[a2] = ss_a1
      c = c + 1

   # remove singleton sheet residues
   c = 4
   cc = len(gap)-4
   while c<cc:
      a0 = gap[c-1]
      a1 = gap[c]
      a2 = gap[c+1]
      if ss[a1] in ['s','S']:
         if ((not ss[a0] in ['s','S']) and
             (not ss[a2] in ['s','S'])):
             ss[a1] = 'L'
      c = c + 1

   # remove sheet residues which aren't next to another sheet 
   c = 4
   cc = len(gap)-4
   while c<cc:
      a1 = gap[c]
      if ss[a1]=='S':
         a1 = gap[c]
         scr_a1 = scr_dict[a1]
         # look for hydrogen bonds to another sheet
         n_a1_atom = n_dict[scr_a1]
         o_a1_atom = o_dict[scr_a1]
         certain = 0
         if n_hb_dict.has_key(n_a1_atom):
            for n_hb_atom in n_hb_dict[n_a1_atom]:
               n_hb_ca_atom=ca_dict[scr_dict[n_hb_atom]]
               if ss[n_hb_ca_atom]=='S':
                  certain = 1
                  break
         if o_hb_dict.has_key(o_a1_atom):
            for o_hb_atom in o_hb_dict[o_a1_atom]:
               o_hb_ca_atom=ca_dict[scr_dict[o_hb_atom]]
               if ss[o_hb_ca_atom]=='S':
                  certain = 1
                  break
         if not certain:
            ss[a1] = 's'
      c = c + 1

   # remove questionable sheet residues
   c = 4
   cc = len(gap)-4
   while c<cc:
      a0 = gap[c-1]
      a1 = gap[c]
      a2 = gap[c+1]
      if ss[a1]=='s':
         if (not ((ss[a0]=='S') and (ss[a2]=='S'))):
            ss[a1] = 'L'
      c = c + 1

   # extend helices if hydrogen bonding requirements are met
   rep_dict = {}
   repeat = 1
   while repeat:
      repeat = 0
      c = 4
      cc = len(gap)-4
      while c<cc:
         a = gap[c]
         if not rep_dict.has_key(a):
            if ss[gap[c+1]]=='H':
               rep_dict[a] = 1
               if ss[a]!='H': # N-terminal end
                  aO = o_dict[scr_dict[a]]
                  ap4N = n_dict[scr_dict[gap[c+4]]]
                  ap3N = n_dict[scr_dict[gap[c+3]]]
                  if hb_dict.has_key((ap4N,aO)) or hb_dict.has_key((ap3N,aO)):
                     ss[a]='H'
                     repeat = 1
                     c = c - 5
                     if c<4: c=4
            if ss[gap[c-1]]=='H':
               a = gap[c]
               if ss[a]!='H': # C-terminal end
                  rep_dict[a] = 1
                  aN = n_dict[scr_dict[a]]
                  am4O = o_dict[scr_dict[gap[c-4]]]
                  am3O = o_dict[scr_dict[gap[c-3]]]
                  if hb_dict.has_key((aN,am4O)) or hb_dict.has_key((aN,am3O)):
                     ss[a]='H'
                     repeat = 1
                     c = c - 5
                     if c<4: c=4
         c = c + 1

   # remove doubleton helices

   c = 4
   cc = len(gap)-5
   while c<cc:
      a0 = gap[c-1]
      a1 = gap[c]
      a2 = gap[c+1]
      a3 = gap[c+2]
      ss_a0 = ss[gap[c-1]]
      ss_a1 = ss[gap[c]]
      ss_a2 = ss[gap[c+1]]
      ss_a3 = ss[gap[c+2]]
      if ss_a1=='H':
         if (ss_a2==ss_a1) and (ss_a0!=ss_a2) and (ss_a2!=ss_a3):
            ss[a1] = 'L'
            ss[a2] = 'L'
      c = c + 1

   # remove totally unreasonable helix and sheet residues

   c = 4
   cc = len(gap)-5
   while c<cc:
      a1 = gap[c]
      ss_a1 = ss[gap[c]]
      if ss_a1=='H':
         if phipsi.has_key(a1):
            (phi,psi) = phipsi[a1]
            if (phi>0) and (phi<150):
               ss[a1] = 'L'
            elif((psi<-120) or (psi>140)):
               ss[a1] = 'L'
      elif ss_a1 in ['S','s']:
         if phipsi.has_key(a1):
            (phi,psi) = phipsi[a1]
            if (phi>45) and (phi<160):
               ss[a1] = 'L'
#            if (psi<-30) and (psi>-150):
            if (psi<-65) and (psi>-150):
               ss[a1] = 'L'
         
      c = c + 1


   for x in range(1,3):
      # remove singleton sheet residues
      c = 4
      cc = len(gap)-4
      while c<cc:
         a0 = gap[c-1]
         a1 = gap[c]
         a2 = gap[c+1]
         if ss[a1] in ['s','S']:
            if ((not ss[a0] in ['s','S']) and
                (not ss[a2] in ['s','S'])):
                ss[a1] = 'L'
         c = c + 1

      # remove sheet residues which aren't next to another sheet 
      c = 4
      cc = len(gap)-4
      while c<cc:
         a1 = gap[c]
         if ss[a1]=='S':
            a1 = gap[c]
            scr_a1 = scr_dict[a1]
            # look for hydrogen bonds to another sheet
            n_a1_atom = n_dict[scr_a1]
            o_a1_atom = o_dict[scr_a1]
            certain = 0
            if n_hb_dict.has_key(n_a1_atom):
               for n_hb_atom in n_hb_dict[n_a1_atom]:
                  n_hb_ca_atom=ca_dict[scr_dict[n_hb_atom]]
                  if ss[n_hb_ca_atom]=='S':
                     certain = 1
                     break
            if o_hb_dict.has_key(o_a1_atom):
               for o_hb_atom in o_hb_dict[o_a1_atom]:
                  o_hb_ca_atom=ca_dict[scr_dict[o_hb_atom]]
                  if ss[o_hb_ca_atom]=='S':
                     certain = 1
                     break
            if not certain:
               ss[a1] = 's'
         c = c + 1

      # remove questionable sheet residues
      c = 4
      cc = len(gap)-4
      while c<cc:
         a0 = gap[c-1]
         a1 = gap[c]
         a2 = gap[c+1]
         if ss[a1]=='s':
            if (not ((ss[a0]=='S') and (ss[a2]=='S'))):
               ss[a1] = 'L'
         c = c + 1

#      lst = ss.keys()
#      lst.sort()
#      for a in lst: print scr_dict[a],ss[a]
      
   # assign protein
   for a in cas:
      if ss[a]=='s':
         ss[a]='S'
      
   cmd.alter(sss1,"ss ='L'")
   for a in cas:
      if ss[a]!='L':
         cmd.alter("(%s`%d)"%a,"ss='%s'"%ss[a])

   cmd.feedback("pop")

   del pymol._ss # IMPORTANT
   cmd.delete(sss1)
   cmd.rebuild(selection,'cartoon')
   #
#   print conn_hash.keys()
   print " util.ss: assignment complete."





from pymol.wizard import Wizard
from pymol import cmd
import pymol

sele_prefix = "_pf_s_"
sele_prefix_len = len(sele_prefix)

dist_prefix = "_pf_d_"


class Pair_fit(Wizard):

   def __init__(self):

      Wizard.__init__(self)
      
      self.memory = 0
      self.n_pair = 0
      self.status = 0 # 0 no atoms selections, 1 atom selected
      self.message = None
      
   def get_panel(self):
      return [
         [ 1, 'Pair Fitting',''],
         [ 2, 'Fit %d Pairs'%self.n_pair,'cmd.get_wizard().fit()'],
         [ 2, 'Delete Last Pair','cmd.get_wizard().remove_last()'],
         [ 2, 'Redraw','cmd.get_wizard().update_dashes()'],         
         [ 2, 'Clear','cmd.get_wizard().clear()'],
         [ 2, 'Done','cmd.set_wizard()'],
         ]

   def cleanup(self):
      self.clear()
      
   def clear(self):
      cmd.delete(sele_prefix+"*")
      cmd.delete(dist_prefix+"*")
      lst = cmd.get_names('selections')
      self.n_pair = 0
      self.status = 0
      self.message = None
      cmd.unpick()
      cmd.refresh_wizard()
      
   def get_prompt(self):
      self.prompt = None
      if self.status==0:
         self.prompt = [ 'Pick the mobile atom...']
      elif self.status==1:
         self.prompt = [ 'Pick the target atom...' ]
      if self.message!=None:
         self.prompt.append(self.message)
      return self.prompt
   
   def set_status(self,status):
      self.status = status
      cmd.refresh_wizard()

   def get_sele_list(self,mode='all'):
      lst = cmd.get_names('selections')
      lst = filter(lambda x:x[0:sele_prefix_len]==sele_prefix,lst)
      lst.sort()
      if mode == 'mobile': # mobile
         lst=filter(lambda x:x[-1:]=='b',lst) 
      elif mode == 'target': # target
         lst=filter(lambda x:x[-1:]=='a',lst)
      return lst
   
   def fit(self):
      # build up the pair-wise list of selections
      cmd.delete(dist_prefix+"*")
      lst = self.get_sele_list()
      c = 0
      args = []
      while 1:
         if not len(lst): break
         a = lst.pop()
         if not len(lst): break
         b = lst.pop()
         args.append(a)
         args.append(b)
      # do the fit
      if len(args):
         cmd.push_undo(args[0])
         dist = apply(cmd.pair_fit,args)
         self.message = "RMS over %d pairs = %5.3f"%(self.n_pair,dist)
         cmd.refresh_wizard()
      self.update_dashes()
                           
   def remove_last(self):
      # build up the pair-wise list of selections
      cmd.delete(dist_prefix+"*")
      lst = self.get_sele_list()
      if len(lst):
         cmd.delete(lst.pop())
         if len(lst):
            cmd.delete(lst.pop())         
         self.n_pair = self.n_pair - 1
      self.update_dashes()
      self.status=0
      cmd.refresh_wizard()
      
   def update_dashes(self):
      cmd.delete(dist_prefix+"*")
      lst = self.get_sele_list()
      c = 0
      while 1:
         if not len(lst): break
         a = lst.pop()
         if not len(lst): break
         b = lst.pop()
         name = dist_prefix+str(c)
         cmd.dist(name,a,b,width=7,length=0.05,gap=0.05)
         cmd.hide('label',name)
         cmd.enable(name)
         c = c + 1

   def check_same_object(self,lst,sele):
      if not len(lst):
         return 1
      else:
         if cmd.count_atoms("((byobj %s) and %s)"%(lst[0],sele),quiet=1):
            return 1
      return 0

   def check_different_object(self,lst,sele):
      if not len(lst):
         return 1
      else:
         if not cmd.count_atoms("((byobj %s) and %s)"%(lst[0],sele),quiet=1):
            return 1
      return 0

   def do_pick(self,bondFlag):
      if bondFlag:
         self.message = "Error: please select an atom, not a bond."
         print self.message
      else:
         if self.status==0:
            lst = self.get_sele_list(mode='mobile')
            if not self.check_same_object(lst,"(pk1)"):
               self.message = "Error: must select an atom in the same object as before."
               print self.message
            else:
               name = sele_prefix + "%02db"%self.n_pair # mobile end in 'b'
               cmd.select(name,"(pk1)")
               cmd.unpick()
               cmd.enable(name)
               self.status = 1
               self.message = None
         elif self.status==1:
            lst = self.get_sele_list(mode='target')
            if not self.check_same_object(lst,"(pk1)"):
               self.message = "Error: must select an atom in the same object as before."
               print self.message
            else:
               lst = self.get_sele_list(mode='mobile')
               if not self.check_different_object(lst,"(pk1)"):
                  self.message = "Error: target atom must be in a distinct object."
                  print self.message
               else:
                  name = sele_prefix + "%02da"%self.n_pair # target end in 'a'
                  cmd.select(name,"(pk1)")
                  cmd.unpick()
                  cmd.enable(name)
                  self.n_pair = self.n_pair + 1
                  self.status = 0
                  self.update_dashes()
            
      cmd.refresh_wizard()



#A* -------------------------------------------------------------------
#B* This file contains source code for the PyMOL computer program
#C* copyright 1998-2000 by Warren Lyford Delano of DeLano Scientific. 
#D* -------------------------------------------------------------------
#E* It is unlawful to modify or remove this copyright notice.
#F* -------------------------------------------------------------------
#G* Please see the accompanying LICENSE file for further information. 
#H* -------------------------------------------------------------------
#I* Additional authors of this source file include:
#-* 
#-* 
#-*
#Z* -------------------------------------------------------------------

import re
import string
import os
import cmd
from cmd import _cmd,lock,unlock,Shortcut,QuietException, \
     _feedback,fb_module,fb_mask, \
     file_ext_re,safe_oname_re, \
     _load
import cmd

from chempy.sdf import SDF,SDFRec

class loadable:
   pdb = 0
   mol = 1
   molstr = 3
   mmod = 4
   xplor = 7
   model = 8
   pdbstr = 9    
   brick = 10    # chempy.brick object
   map = 11      # chempy.map object
   callback = 12 # pymol callback obejct
   cgo = 13      # compiled graphic object
   r3d = 14      # r3d, only used within cmd.py
   xyz = 15      # xyz, tinker format
   sdf = 16      # sdf, only used within cmd.py
   cc1 = 17      # cc1 and cc2, only used within cmd.py
   ccp4 = 18     # CCP4 map, under development
   pmo = 19      # pmo, experimental molecular object format
   
loadable_sc = Shortcut(loadable.__dict__.keys()) 

def load_object(type,object,name,state=0,finish=1,discrete=0):
      # assume first argument is the object type (numeric)
   '''
DESCRIPTION

   "load_object" is a general developer function for loading Python objects
   into PyMOL.

PYMOL API

   cmd.load_object(type,object,name,state=0,finish=1,discrete=0)

   NOTE type is one one of the numberic cmd.loadable types
   '''
   r = 1
   try:
      lock()   
      r = _cmd.load_object(str(name),object,int(state)-1,
                              int(type),int(finish),int(discrete))
   finally:
      unlock()
   return r
   
def load_brick(*arg):
   '''
Temporary routine for GAMESS-UK project.
'''
   lst = [loadable.brick]
   lst.extend(list(arg))
   return apply(load_object,lst)

def load_map(*arg):
   '''
Temporary routine for the Phenix project.
'''
   
   lst = [loadable.map]
   lst.extend(list(arg))
   return apply(load_object,lst)

def load_callback(*arg):
   '''
DESCRIPTION

   "load_callback" is used to load a generic Python callback object.
   These objects are called every time the screen is updated and can be used
   to trigger OpenGL rendering calls (such as with PyOpenGL).

PYMOL API

   cmd.load_callback(object,name,state,finish,discrete)
   
'''
   
   lst = [loadable.callback]
   lst.extend(list(arg))
   return apply(load_object,lst)

def load_cgo(*arg):
   '''
DESCRIPTION

   "load_cgo" is used to load a compiled graphics object, which is
   actually a list of floating point numbers built using the constants
   in the $PYMOL_PATH/modules/pymol/cgo.py file.

PYMOL API

   cmd.load_cgo(object,name,state,finish,discrete)
   
'''
   
   lst = [loadable.cgo]
   lst.extend(list(arg))
   return apply(load_object,lst)

def load_model(*arg,**kw):
   '''
DESCRIPTION
  
   "load_model" reads a ChemPy model into an object

PYMOL API
  
   cmd.load_model(model, object [,state [,finish [,discrete ]]])
   '''
   lst = [loadable.model]
   lst.extend(list(arg))
   return apply(load_object,lst,kw)

def load(filename,object='',state=0,format='',finish=1,discrete=0):
   '''
DESCRIPTION
  
   "load" reads several file formats.  The file extension is used to
   determine the format.  PDB files must end in ".pdb", MOL files must
   end in ".mol", Macromodel files must end in ".mmod", XPLOR
   maps must end in ".xplor", CCP4 maps must end in ".ccp4",
   Raster3D input (Molscript output) must end in ".r3d".

   Pickled ChemPy models with a ".pkl" can also be directly read.
 
   If an object is specified, then the file is load into that object.
   Otherwise, an object is created with the same name as the file
   prefix.
 
USAGE
 
   load filename [,object [,state [,format [,finish [,discrete ]]]]]
 
PYMOL API
  
   cmd.load( filename [,object [,state [,format [,finish [,discrete ]]]]]

NOTES

   You can override the file extension by giving a format string:

   'pdb' : PDB,  'mmod' : Macromodel, 'xyz' : Tinker, 'cc1' : ChemDraw3D  
   'mol' : MDL MOL-file, 'sdf' : MDL SD-file
   'xplor' : X-PLOR/CNS map, 'ccp4' : CCP4 map,
   'callback' : PyMOL Callback object (PyOpenGL)
   'cgo' : compressed graphics object (list of floats)

SEE ALSO

   save
   '''
   r = 1
   try:
      lock()
      type = format
      ftype = 0
      state = int(state)
      finish = int(finish)
      discrete = int(discrete)
      fname = filename
      fname = os.path.expanduser(fname)
      fname = os.path.expandvars(fname)

      if not len(str(type)):
         # determine file type if possible
         if re.search("\.pdb$|\.ent$",filename,re.I):
            ftype = loadable.pdb
         elif re.search("\.mol$",filename,re.I):
            ftype = loadable.mol
         elif re.search("\.mmod$|\.mmd$|\.dat$|\.out$",filename,re.I):
            ftype = loadable.mmod
         elif re.search("\.xplor$",filename,re.I):
            ftype = loadable.xplor
         elif re.search("\.ccp4$",filename,re.I):
            ftype = loadable.ccp4
         elif re.search("\.pkl$",filename,re.I):
            ftype = loadable.model
         elif re.search("\.r3d$",filename,re.I):
            ftype = loadable.r3d
         elif re.search("\.xyz$",filename,re.I):
            ftype = loadable.xyz
         elif re.search("\.cc1$|\.cc2$",filename,re.I):
            ftype = loadable.cc1
         elif re.search("\.xyz_[0-9]*$",filename,re.I):
            ftype = loadable.xyz
         elif re.search("\.sdf$",filename,re.I):
            ftype = loadable.sdf
         elif re.search("\.pmo$",filename,re.I):
            ftype = loadable.pmo
         else:
            ftype = loadable.pdb # default is PDB
      elif cmd.is_string(type):
         try:
            ftype = int(type)
         except:
            type = loadable_sc.auto_err(type,'file type')
            if hasattr(loadable,type):
               ftype = getattr(loadable,type)
            else:
               print "Error: unknown type '%s'",type
               raise QuietException
      else:
         ftype = int(type)
         
# get object name
      if len(str(object))==0:
         oname = re.sub(r".*\/|.*\\","",filename) # strip path
         oname = file_ext_re.sub("",oname) # strip extension
         oname = safe_oname_re.sub("_",oname)
         if not len(oname): # safety
            oname = 'obj01'
      else:
         oname = string.strip(object)

# special handling of sdf files
      if ftype == loadable.sdf:
         ftype = loadable.molstr
         sdf = SDF(fname)
         while 1:
            rec = sdf.read()
            if not rec: break
            r = _load(oname,string.join(rec.get('MOL'),''),state,ftype,0,1)
         del sdf
         _cmd.finish_object(str(oname))
         _cmd.do("zoom (%s)"%oname) 
         ftype = -1
         
# standard file handling

      if ftype>=0:
         r = _load(oname,fname,state,ftype,finish,discrete)
   finally:
      unlock()
   return r

def read_molstr(molstr,name,state=0,finish=1,discrete=1):
   '''
DESCRIPTION
  
   "read_molstr" reads an MDL MOL format file as a string
   
PYMOL API ONLY
 
   cmd.read_molstr( string molstr, string name, int state=0,
      int finish=1, int discrete=1 )

NOTES

   "state" is a 1-based state index for the object, or 0 to append.

   "finish" is a flag (0 or 1) which can be set to zero to improve
   performance when loading large numbers of objects, but you must
   call "finish_object" when you are done.

   "discrete" is a flag (0 or 1) which tells PyMOL that there will be
   no overlapping atoms in the file being loaded.  "discrete"
   objects save memory but can not be edited.
   '''
   r = 1
   try:
      lock()
      r = _cmd.load(str(name),str(molstr),int(state)-1,
                    loadable.molstr,int(finish),int(discrete))
   finally:
      unlock()
   return r

def read_mmodstr(*arg):
   '''
DESCRIPTION

   "read_mmodstr" reads a macromodel format structure from a Python
   string.

'''
   r = 1
   try:
      lock()   
      ftype = 6
      if len(arg)==2:
         oname = string.strip(arg[1])
         r = _cmd.load(str(oname),arg[0],-1,int(ftype),1,1)
      elif len(arg)==3:
         oname = string.strip(arg[1])
         r = _cmd.load(str(oname),arg[0],int(arg[2])-1,int(ftype),1,1)
      else:
         print "argument error."
   finally:
      unlock()
   return r

def read_pdbstr(pdb,name,state=0,finish=1,discrete=0):
   '''
DESCRIPTION
  
   "read_pdbstr" in an API-only function which reads a pdb file from a
   Python string.  This feature can be used to load or update
   structures into PyMOL without involving any temporary files.
   
PYMOL API ONLY
 
   cmd.read_pdbstr( string pdb-content, string object name 
      [ ,int state [ ,int finish [ ,int discrete ] ] ] )

NOTES

   "state" is a 1-based state index for the object.

   "finish" is a flag (0 or 1) which can be set to zero to improve
   performance when loading large numbers of objects, but you must
   call "finish_object" when you are done.

   "discrete" is a flag (0 or 1) which tells PyMOL that there will be
   no overlapping atoms in the PDB files being loaded.  "discrete"
   objects save memory but can not be edited.
'''
   r = 1
   try:
      lock()   
      ftype = loadable.pdbstr
      oname = string.strip(str(name))
      r = _cmd.load(str(oname),pdb,int(state)-1,int(ftype),
                       int(finish),int(discrete))
   finally:
      unlock()
   return r

def finish_object(name):
   '''
DESCRIPTION

   "finish_object" is used in cases where many individual states are
   being loaded and it is advantageos to avoid processing them until
   all states have been loaded into RAM.  This function should always
   be called after loading an object with the finish flag set to zero.

PYMOL API

   cmd.finish(string name)

   "name" should be the name of the object
   '''
   r = 1
   try:
      lock()   
      r = _cmd.finish_object(name)
   finally:
      unlock()
   return r



import sys
import string

f = sys.stdin
g = sys.stdout

echo = 0
while 1:
   l = f.readline()
   if not l: break
   ll=string.strip(l)
   if ll=='BEGIN-LOG':
      echo = 1
   elif ll=='END-LOG':
      echo = 0
   elif echo:
      g.write(l)



try:
   import _champ
except ImportError:
   print " Error: unable to import architecture-dependent _champ module."

# okay, this module is going to take some planning, since Champ
# has the potential to evolve into a full-blown
# chemical informatics system 

# what do we need to be prepared to handle?
#
# - single structures
# - single patterns
# - lists of patterns
# - lists of compounds
# - dictionaries of compounds (key/value)
# - molecules in other class formats (chempy models, etc.)
# - subsets of protein structures
# - scripts which apply rules according to patterns
# - synchronized chempy models

# what does the architecture-dependent modules need to support?

# - Smiles/SMARTS/etc. conversion (bidirectional)
# - ChemPy model conversion (bidirectional)
# - modification recipe for extant chempy models?...

class Champ:

   def __init__(self):
      '''
      creates a new champ object, which contains
      a collect of comparable, searchable patterns 
      '''
      self._champ = _champ.new()

   def insert_pattern_string(self,pattern):
      '''
      adds a new pattern, and
      returns the index to this pattern
      '''
      (e,r) = _champ.insert_pattern_string(self._champ,str(pattern))
      if e: raise RuntimeError
      return r

   def pattern_free(self,index):
      '''
      frees a pattern
      '''
      (e,r) = _champ.pattern_free(self._champ,int(index))
      if e: raise RuntimeError
      return r

   def pattern_clear_tags(self,index):
      '''
      resets a pattern\'s tags
      '''
      (e,r) = _champ.pattern_clear_tags(self._champ,int(index))
      if e: raise RuntimeError
      return r
      
   def get_pattern_string(self,index):
      '''
      retrieves the smiles string for a given pattern index
      '''
      (e,r) = _champ.get_pattern_string(self._champ,int(index))
      if e: raise RuntimeError
      return r
   
   def insert_model(self,model):
      '''
      inserts the pattern from a ChemPy model pattern and returns the index
      '''
      (e,r) = _champ.insert_model(self._champ,model)
      if e: raise RuntimeError
      return r

   def list_prepend_pattern_strings(self,handle,pattern):
      '''
      adds pattern string to the top of a list
      and returns the new list identifier
      '''
      (e,r) = _champ.list_prepend_pattern_strings(self._champ,int(handle),pattern)
      if e: raise RuntimeError
      return r

   def list_new(self):
      '''
      returns a new list handle
      '''
      (e,r) = _champ.list_new(self._champ)
      if e: raise RuntimeError
      return r

   def list_free(self,handle,purge=1):
      '''
      destroys a list and associated patterns (if purge = 1)
      '''
      (e,r) = _champ.list_free(self._champ,int(handle),int(purge))
      if e: raise RuntimeError
      return r

   def list_get_pattern_indices(self,handle):
      '''
      returns pattern indices in a list as a list
      '''
      (e,r) = _champ.list_get_pattern_list(self._champ,int(handle))
      if e: raise RuntimeError
      return r

   def list_get_pattern_strings(self,handle):
      '''
      returns list of smiles strings contained in a list
      '''
      (e,r) = _champ.list_get_pattern_strings(self._champ,int(handle))
      if e: raise RuntimeError
      return r

   def match_1v1_b(self,pattern,target): # boolean
      '''
      returns whether or not pattern can be found in target
      as a boolean result
      '''
      (e,r) = _champ.match_1v1_b(self._champ,
                           int(pattern),int(target))
      if e: raise RuntimeError
      return r

   def match_1v1_n(self,pattern,target,limit,tag=0):
      '''
      returns tag list for single pattern on target
      '''
      (e,r) = _champ.match_1v1_n(self._champ,
                           int(pattern),int(target),
                                 int(limit),int(tag))
      if e: raise RuntimeError
      return r

   def match_1v1_map(self,pattern,target,limit): # boolean
      '''
      returns mappings (if any) between two patterns
      '''
      (e,r) = _champ.match_1v1_map(self._champ,
                           int(pattern),int(target),int(limit))
      if e: raise RuntimeError
      return r

   def match_Nv1_n(self,pattern,target,limit,tag):
      '''
      returns tag list for list of patterns onto a target
      '''
      (e,r) = _champ.match_Nv1_n(self._champ,
                           int(pattern),int(target),
                                 int(limit),int(tag))
      if e: raise RuntimeError
      return r

      
   def match_1vN_n(self,pattern,handle):
      '''
      returns count of how many times pattern occurs in list
      '''
      (e,r) = _champ.match_1vN_n(self._champ,
                           int(pattern),int(handle))
      if e: raise RuntimeError
      return r

   def memory_dump(self):
      '''
      dump bulk memory information
      '''
      (e,r) = _champ.memory_dump(self._champ)
      if e: raise RuntimeError
      return r

   def pattern_get_cycle(self,index):
      '''
      debugging routine
      '''
      (e,r) = _champ.pattern_get_cycle(self._champ,int(index))
      if e: raise RuntimeError
      return r

   def pattern_get_class(self,index):
      '''
      debugging routine
      '''
      (e,r) = _champ.pattern_get_class(self._champ,int(index))
      if e: raise RuntimeError
      return r

   def pattern_get_codes(self,index):
      '''
      debugging routine
      '''
      (e,r) = _champ.pattern_get_codes(self._champ,int(index))
      if e: raise RuntimeError
      return r

   def pattern_get_tags(self,index):
      '''
      get atomic tags
      '''
      (e,r) = _champ.pattern_get_tags(self._champ,int(index))
      if e: raise RuntimeError
      return r

   def pattern_get_atom_symbols(self,index):
      '''
      debugging routine
      '''
      (e,r) = _champ.pattern_get_atom_symbols(self._champ,int(index))
      if e: raise RuntimeError
      return r

   def pattern_dump(self,index):
      '''
      debugging routine
      '''
      (e,r) = _champ.pattern_dump(self._champ,int(index))
      if e: raise RuntimeError
      return r





#A* -------------------------------------------------------------------
#B* This file contains source code for the PyMOL computer program
#C* copyright 1998-2000 by Warren Lyford Delano of DeLano Scientific. 
#D* -------------------------------------------------------------------
#E* It is unlawful to modify or remove this copyright notice.
#F* -------------------------------------------------------------------
#G* Please see the accompanying LICENSE file for further information. 
#H* -------------------------------------------------------------------
#I* Additional authors of this source file include:
#-* 
#-* 
#-*
#Z* -------------------------------------------------------------------

import sglite

# xray.py 
# This section contains python code for supporting
# x-ray crystallography functions

def sg_sym_to_mat_list(sgsymbol):
   try:
      Symbols_Inp = sglite.SgSymbolLookup(sgsymbol)
      if(Symbols_Inp):
         HallSymbol = Symbols_Inp['Hall']
         SgOps = sglite.SgOps(HallSymbol)
         nLTr = SgOps.get_nLTr()
         fInv = SgOps.get_fInv()
         nSMx = SgOps.get_nSMx()
         result = []
         rb = float(sglite.SRBF)
         tb = float(sglite.STBF)
         for iLTr in xrange(nLTr):
            for iInv in xrange(fInv):
               for iSMx in xrange(nSMx):
                  Mx = SgOps.getLISMx(iLTr, iInv, iSMx, +1)
                  result.append([[ Mx[0]/rb, Mx[1]/rb, Mx[2]/rb, Mx[9 ]/tb],
                                 [ Mx[3]/rb, Mx[4]/rb, Mx[5]/rb, Mx[10]/tb],
                                 [ Mx[6]/rb, Mx[7]/rb, Mx[8]/rb, Mx[11]/tb],
                                 [        0,        0,        0,         1]] )                  
   except:
      print " pmx: unrecognized space group symbol '"+sgsymbol+"'"
      result = None
   return result




#A* -------------------------------------------------------------------
#B* This file contains source code for the PyMOL computer program
#C* copyright 1998-2000 by Warren Lyford Delano of DeLano Scientific. 
#D* -------------------------------------------------------------------
#E* It is unlawful to modify or remove this copyright notice.
#F* -------------------------------------------------------------------
#G* Please see the accompanying LICENSE file for further information. 
#H* -------------------------------------------------------------------
#I* Additional authors of this source file include:
#-* 
#-* 
#-*
#Z* -------------------------------------------------------------------

# cmd.py 
# Python interface module for PyMol
#
# **This is the only module which should be/need be imported by 
# ** PyMol API Based Programs

# NOTE: this cmd module has grown absurdly huge, therefore expect this
# file to be broken down sometime after the 0.51 release into more
# manageable pieces.  Note that cmd.<whatever> will still work -- the
# symbols will be mapped into this namespace even after the code modules
# are separated.

# NEW CALL RETURN CONVENTIONS for _cmd.so C-layer
#
# (1) Calls into C (_cmd) should return results/status and print errors
#     and feedback (according to mask) BUT NEVER RAISE EXCEPTIONS.
## (2) In the absence of an expected return value, truth applies:
#        Success is 1, true 
#        Failure is 0, false. None, NULL
#
#     NOTE: if you need more informative failure, then code and
#           document an exception to this rule for your functions!
#
# (3) If _cmd produces a specific return result, be sure to include an
#     error result as one of the possibilities outside the range of the
#     expected return value.  For example, a negative distance or count
#
# (4) cmd.py API wrappers can then raise exceptions based on truth
#     and should return truth for success or None for failure
#     (if no exception was raised)
#

import re
import _cmd
import string
import thread
import threading
import types
import pymol
import os
import parsing
import sys

from shortcut import Shortcut

from chempy import io

#######################################################################
# symbols for early export
#######################################################################

file_ext_re= re.compile(string.join([
   "\.pdb$|\.ent$|\.mol$|",
   r"\.PDB$|\.ENT$|\.MOL$|",
   r"\.mmod$|\.mmd$|\.dat$|\.out$|",
   r"\.MMOD$|\.MMD$|\.DAT$|\.OUT$|",
   r"\.xplor$|\.pkl$|\.sdf$|", 
   r"\.XPLOR$|\.PKL$|\.SDF$|",                        
   r"\.r3d$|\.xyz$|\.xyz_[0-9]*$|", 
   r"\.R3D$|\.XYZ$|\.XYZ_[0-9]*$|",
   r"\.cc1$|\.cc2$|", # ChemDraw 3D
   r"\.CC1$|\.CC2$|",
   r"\.pmo$|", # Experimental molecular object format
   r"\.PMO$|",
   r"\.ccp4$|\.CCP4$" # CCP4   
   ],''))

safe_oname_re = re.compile(r"\+|\(|\)|\||\&|\!|\,")  # quash reserved characters
   
QuietException = parsing.QuietException

#--------------------------------------------------------------------
# shortcuts...

toggle_dict = {'on':1,'off':0,'1':1,'0':0}
toggle_sc = Shortcut(toggle_dict.keys())

stereo_dict = {'on':1,'off':0,'1':1,'0':0,'swap':-1,
               'crosseye':2,'quadbuffer':3} #,'walleye':3}
stereo_sc = Shortcut(stereo_dict.keys())

repres = {
   'everything'    : -1,
   'sticks'        : 0,
   'spheres'       : 1,
   'surface'       : 2,
   'labels'        : 3,
   'nb_spheres'    : 4,
   'cartoon'       : 5,
   'ribbon'        : 6,
   'lines'         : 7,
   'mesh'          : 8,
   'dots'          : 9,
   'dashes'        :10,
   'nonbonded'     :11,
   'cell'          :12,
   'cgo'           :13,
   'callback'      :14,
   'extent'        :15,   
}
repres_sc = Shortcut(repres.keys())

def null_function():
   pass

#--------------------------------------------------------------------
# convenient type-checking

def is_string(obj):
   return isinstance(obj,types.StringType)

def is_list(obj):
   return isinstance(obj,types.ListType)

#--------------------------------------------------------------------
# locks and threading

# the following lock is used by both C and Python to insure that no more than
# one active thread enters PyMOL at a given time. 

lock_api = pymol.lock_api
lock_api_c = pymol.lock_api_c

def lock_c(): # INTERNAL
   lock_api_c.acquire(1)

def unlock_c(): # INTERNAL
   lock_api_c.release()

def lock(): # INTERNAL
   lock_api.acquire(1) 

def lock_attempt(): # INTERNAL
   return lock_api.acquire(blocking=0)

def unlock(): # INTERNAL
   if (thread.get_ident() == pymol.glutThread):
      lock_api.release()
      _cmd.flush_now()
   else:
      lock_api.release()
      if _cmd.wait_queue(): # commands waiting to be executed?
         w = 0.0025 # NOTE: affects API perf. for "do" and delayed-exec
         while 1:
            e = threading.Event() # using this for portable delay
            e.wait(w)
            del e
            if not _cmd.wait_queue():
               break
            if w > 0.1: # wait up 0.2 sec max for PyMOL to flush queue
               if _feedback(fb_module.cmd,fb_mask.debugging):
                  fb_debug.write("Debug: avoiding possible dead-lock?\n")
               break
            w = w * 2 # wait twice as long each time until flushed

def is_glut_thread(): # internal
   if thread.get_ident() == pymol.glutThread:
      return 1
   else:
      return 0

#--------------------------------------------------------------------
# Feedback

class fb_action:
   set = 0
   enable = 1
   disable = 2
   push = 3
   pop = 4
   
fb_action_sc = Shortcut(fb_action.__dict__.keys())

class fb_module:

# This first set represents internal C systems

   all                       =0
   isomesh                   =1
   map                       =2
   matrix                    =3
   mypng                     =4
   triangle                  =5
   match                     =6
   raw                       =7
   isosurface                =8
   
   feedback                  =12
   scene                     =13
   threads                   =14  
   symmetry                  =15
   ray                       =16
   setting                   =17
   object                    =18
   ortho                     =19
   movie                     =20
   python                    =21
   extrude                   =22
   rep                       =23
   shaker                    =24
   
   coordset                  =25
   distset                   =26

   objectmolecule            =30
   objectmap                 =31
   objectmesh                =32
   objectdist                =33 
   objectcgo                 =34
   objectcallback            =35
   objectsurface             =36
   
   repwirebond               =45
   repcylbond                =46
   replabel                  =47
   repsphere                 =49
   repsurface                =50
   repmesh                   =51
   repdot                    =52
   repnonbonded              =53
   repnonbondedsphere        =54
   repdistdash               =55
   repdistlabel              =56
   repribbon                 =57
   repcartoon                =58
   sculpt                    =59
   
   executive                 =70
   selector                  =71
   editor                    =72

   export                    =75
   ccmd                      =76
   api                       =77   
   
   main                      =80  

# This second set, with negative indices
# represent python level systems

   parser                    =-1
   cmd                       =-2
   
fb_module_sc = Shortcut(fb_module.__dict__.keys())

class fb_mask:
   results =             0x01
   errors =              0x02
   actions =             0x04
   warnings =            0x08
   details =             0x10
   blather =             0x20
   debugging =           0x80
   everything =          0xFF
   
fb_mask_sc = Shortcut(fb_mask.__dict__.keys())

fb_dict ={}

for a in fb_module.__dict__.keys():
   vl = getattr(fb_module,a)
   if vl<0:
      fb_dict[vl] = 0x1F # default mask

fb_debug = sys.stderr # can redirect python debugging output elsewhere if desred...
   
def feedback(action="?",module="?",mask="?"):
   '''
DESCRIPTION

   "feedback" allows you to control what and how much text is output
   from PyMOL.

USAGE
   
   feedback action,module,mask

   action is one of ['set','enable','disable']
   module is a space-separated list of strings or simply "all"
   mask is a space-separated list of strings or simply "everything"

NOTES:

   "feedback" alone will print a list of the available choices
   
PYMOL API

   cmd.feedback(string action,string module,string mask)
   
EXAMPLES

   feedback enable, all , debugging
   feedback disable, selector, warnings actions
   feedback enable, main, blather
'''
   r = None

   # validate action

   if action=="?":
      print " feedback: available actions: set, enable, disable"
      act_kee = 0
   else:
      act_kee = fb_action_sc.interpret(action)
      if act_kee == None:
         print "Error: invalid feedback action '%s'."%action
         if _raising():
            raise QuietException
         else:
            return None
      elif not is_string(act_kee):
         print "Error: ambiguous feedback action '%s'."%action
         print action_amb
         if _raising():
            raise QuietException
         else:
            return None
      act_int = int(getattr(fb_action,act_kee))

   if (act_int<3) and ("?" in [action,module,mask]):
      if module=="?":
         print " feedback: available modules:"
         for a in fb_module.__dict__.keys():
            if a[0]!='_':
               print "   ",a
      if mask=="?":
         print " feedback: available masks:"
         for a in fb_mask.__dict__.keys():
            if a[0]!='_':
               print "   ",a
   else:
      if (act_int>=3):
         module='all'
         mask='everything'
            
      # validate and combine masks
      
      mask_int = 0
      mask_lst = string.split(mask)
      for mask in mask_lst:
         mask_kee = fb_mask_sc.interpret(mask)
         if mask_kee == None:
            print "Error: invalid feedback mask '%s'."%mask
            if _raising(): raise QuietException
            else: return None
         elif not is_string(mask_kee):
            print "Error: ambiguous feedback mask '%s'."%mask
            if _raising(): raise QuietException
            else: return None
         mask_int = int(getattr(fb_mask,mask_kee))

      # validate and iterate modules
      
      mod_lst = string.split(module)
      for module in mod_lst:
         mod_kee = fb_module_sc.interpret(module)
         if mod_kee == None:
            print "Error: invalid feedback module '%s'."%module
            if _raising(): raise QuietException
            else: return None
         elif not is_string(mod_kee):
            print "Error: ambiguous feedback module '%s'."%module
            if _raising(): raise QuietException
            else: return None
         mod_int = int(getattr(fb_module,mod_kee))
         if mod_int>=0:
            try:
               lock()
               r = _cmd.set_feedback(act_int,mod_int,mask_int)
            finally:
               unlock()
         if mod_int<=0:
            if mod_int:
               if act_int==0:
                  fb_dict[mod_int] = mask_int
               elif act_int==1:
                  fb_dict[mod_int] = fb_dict[mod_int] | mask_int
               elif act_int==2:
                  fb_dict[mod_int] = fb_dict[mod_int] & ( 0xFF - mask_int )
            else:
               for mod_int in fb_dict.keys():
                  if act_int==0:
                     fb_dict[mod_int] = mask_int
                  elif act_int==1:
                     fb_dict[mod_int] = fb_dict[mod_int] | mask_int
                  elif act_int==2:
                     fb_dict[mod_int] = fb_dict[mod_int] & ( 0xFF - mask_int )
            if _feedback(fb_module.feedback,fb_mask.debugging):
                sys.stderr.write(" feedback: mode %d on %d mask %d\n"%(
                   act_int,mod_int,mask_int))
   return r

#--------------------------------------------------------------------
# internal API routines

# status reporting

def _feedback(module,mask): # feedback test routine
   r = 0
   module = int(module)
   mask = int(mask)
   if module>0:
      try:
         lock()
         r = _cmd.feedback(module,mask)
      finally:
         unlock()
   else:
      if fb_dict.has_key(module):
         r = fb_dict[module]&mask
   return r

# movie rendering

def _mpng(*arg): # INTERNAL
   try:
      lock()   
      fname = arg[0]
      if re.search("\.png$",fname):
         fname = re.sub("\.png$","",fname)
      fname = os.path.expanduser(fname)
      fname = os.path.expandvars(fname)
      r = _cmd.mpng_(str(fname),arg[1],arg[2])
   finally:
      unlock()
   return r

# loading

def _load(oname,finfo,state,ftype,finish,discrete):
   # caller must already hold API lock
   # NOTE: state index assumes 1-based state
   r = 1
   if ftype not in (loadable.model,loadable.brick):
      if ftype == loadable.r3d:
         import cgo
         obj = cgo.from_r3d(finfo)
         if obj:
            _cmd.load_object(str(oname),obj,int(state)-1,loadable.cgo,
                             int(finish),int(discrete))
         else:
            print " load: couldn't load raster3d file."
      elif ftype == loadable.cc1: # ChemDraw 3D
         obj = io.cc1.fromFile(finfo)
         if obj:
            _cmd.load_object(str(oname),obj,int(state)-1,loadable.model,
                             int(finish),int(discrete))            
      else:
         r = _cmd.load(str(oname),finfo,int(state)-1,int(ftype),
                       int(finish),int(discrete))
   else:
      try:
         x = io.pkl.fromFile(finfo)
         if isinstance(x,types.ListType) or isinstance(x,types.TupleType):
            for a in x:
               r = _cmd.load_object(str(oname),a,int(state)-1,
                                    int(ftype),0,int(discrete))
               if(state>0):
                  state = state + 1
            _cmd.finish_object(str(oname))
         else:
            r = _cmd.load_object(str(oname),x,
                                 int(state)-1,int(ftype),
                                 int(finish),int(discrete))
      except:
         print 'Error: can not load file "%s"' % finfo
   return r

# function keys and other specials

def _special(k,x,y): # INTERNAL (invoked when special key is pressed)
   k=int(k)
   if special.has_key(k):
      if special[k][1]:
         apply(special[k][1],special[k][2],special[k][3])
   return None

# control keys

def _ctrl(k):
   if ctrl.has_key(k):
      ck = ctrl[k]
      if ck[0]!=None:
         apply(ck[0],ck[1],ck[2])
   return None

# alt keys

def _alt(k):
   if alt.has_key(k):
      ak=alt[k]
      if ak[0]!=None:
         apply(ak[0],ak[1],ak[2])
   return None

# writing PNG files (thread-unsafe)

def _png(a): # INTERNAL - can only be safely called by GLUT thread 
   try:
      lock()   
      fname = a
      if not re.search("\.png$",fname):
         fname = fname +".png"
      fname = os.path.expanduser(fname)
      fname = os.path.expandvars(fname)         
      r = _cmd.png(str(fname))
   finally:
      unlock()
   return r

# quitting (thread-specific)

def _quit():
   try:
      lock()
      r = _cmd.quit()
   finally:
      unlock()
   return r

# screen redraws (thread-specific)

def _refresh(swap_buffers=1):  # Only call with GLUT thread!
   try:
      lock()
      if thread.get_ident() == pymol.glutThread:
         if swap_buffers:
            r = _cmd.refresh_now()
         else:
            r = _cmd.refresh()
      else:
         print "Error: Ignoring an unsafe call to cmd._refresh"
   finally:      unlock()
   return r

# stereo (platform dependent )

def _sgi_stereo(flag): # SGI-SPECIFIC - bad bad bad
   
   if os.path.exists("/usr/gfx/setmon"):
      if flag:
         os.system("/usr/gfx/setmon -n 1024x768_96s")
      else:
         os.system("/usr/gfx/setmon -n 72hz")
      
# color alias interpretation

def _interpret_color(color):
   _validate_color_sc()
   new_color = color_sc.interpret(color)
   if new_color:
      if is_string(new_color):
         return new_color
      else:
         color_sc.auto_err(color,'color')
   else:
      return color

def _validate_color_sc():
   global color_sc
   if color_sc == None: # update color shortcuts if needed
      lst = get_color_indices()
      color_sc = Shortcut(map(lambda x:x[0],lst))
      color_dict = {}
      for a in lst: color_dict[a[0]]=a[1]

def _invalidate_color_sc():
   global color_sc
   color_sc = None

def _get_color_sc():
   _validate_color_sc()
   return color_sc

def _get_feedback(): # INTERNAL
   l = []
   if lock_attempt():
      try:
         r = _cmd.get_feedback()
         while r:
            l.append(r)
            r = _cmd.get_feedback()
      finally:
         unlock()
   return l
get_feedback = _get_feedback # for legacy compatibility

# testing tools

# for comparing floating point numbers calculated using
# different FPUs

def _dump_floats(lst,format="%7.3f",cnt=9):
   c = cnt
   for a in lst:
      print format%a,
      c = c -1
      if c<=0:
         print
         c=cnt
   if c!=cnt:
      print
   
# HUH?
def _adjust_coord(a,i,x):
   a.coord[i]=a.coord[i]+x
   return None

def _raising():
   return get_setting_legacy("raise_exceptions")

#######################################################################
# now import modules which depend on the above
#######################################################################

import editor

#######################################################################
# cmd module functions...
#######################################################################

def ready(): # INTERNAL
   return _cmd.ready()

def setup_global_locks(): # INTERNAL, OBSOLETE?
   pass
   
def null():
   pass

# for extending the language

def extend(name,function):
   '''
DESCRIPTION

   "extend" is an API-only function which binds a new external
   function as a command into the PyMOL scripting language.

PYMOL API

   cmd.extend(string name,function function)
   
PYTHON EXAMPLE

   def foo(moo=2): print moo
   cmd.extend('foo',foo)

   The following would now be valid within PyMOL:

   foo
   foo 3
   foo moo=5

SEE ALSO

   alias, api
   '''
   keyword[name] = [function, 0,0,',',parsing.STRICT]
   kwhash.append(name)
   

# for aliasing compound commands to a single keyword

def alias(name,command):
   '''
DESCRIPTION

   "alias" allows you to bind a commonly used command to a single word

USAGE
   
   alias name, command-sequence

PYMOL API

   cmd.alias(string name,string command)
   
EXAMPLES

   alias go,load "test.pdb"; zoom (i;500); show sticks,(i;500 a;4)
   go

SEE ALSO

   extend
   '''
   keyword[name] = [eval("lambda :do('''%s ''')"%command), 0,0,',',parsing.STRICT]
   kwhash.append(name)

def write_html_ref(file):
   lst = cmd.globals()
   f=open(file,'w')
   head = 'H2'
   f.write("<HTML><BODY><H1>Reference</H1>")
   kees = lst.keys()
   kees.sort()
   for a in kees:
      if hasattr(lst[a],'__doc__'):
         if a[0:1]!='_' and (a not in ['string','thread',
                                       'setup_global_locks',
                                       'real_system','sys','imp','glob'
                                       ]):
            doc = lst[a].__doc__
            if is_string(doc):
               if len(doc):
                  doc = string.strip(doc)
                  doc = string.replace(doc,"<","&lt;")
                  f.write("<HR SIZE=1><%s>"%head+a+"</%s>\n"%head)
                  f.write("<PRE>"+string.strip(doc)+"\n\n</PRE>")
   f.write("</BODY></HTML>")
   f.close()

def dummy(*arg):
   '''
DESCRIPTION

   This is a dummy function which simply returns None.
   Don't you wish your life was so easy?
   '''
   #'
   return None

#####################################################################
# Here is where the PyMOL Command Language and API are built.
#####################################################################


# first we need to import a set of symbols into this module's local
# namespace

#--------------------------------------------------------------------
from importing import \
     load,               \
     load_brick,         \
     load_callback,      \
     load_cgo,           \
     load_map,           \
     load_model,         \
     load_object,        \
     read_mmodstr,       \
     read_molstr,        \
     read_pdbstr,        \
     loadable    

#--------------------------------------------------------------------
from creating import \
     copy,               \
     create,             \
     fragment,           \
     isodot,             \
     isomesh,            \
     isosurface,         \
     symexp,             \
     map_new

#--------------------------------------------------------------------
from commanding import \
     cls,                \
     delete,             \
     do,                 \
     log,                \
     log_close,          \
     log_open,           \
     quit,               \
     resume,             \
     splash,             \
     sync

#--------------------------------------------------------------------
import controlling
from controlling import \
     button,             \
     config_mouse,       \
     mouse,              \
     mask,               \
     set_key,            \
     unmask,             \
     edit_mode

#--------------------------------------------------------------------
from querying import \
     count_atoms,        \
     count_frames,       \
     count_states,       \
     dist,               \
     distance,           \
     export_dots,        \
     find_pairs,         \
     get_area,           \
     get_color_indices,  \
     get_color_tuple,    \
     get_dihedral,       \
     get_extent,         \
     get_model,          \
     get_names,          \
     get_names_of_type,  \
     get_phipsi,         \
     get_position,       \
     get_povray,         \
     get_renderer,       \
     get_title,          \
     get_type,           \
     id_atom,            \
     identify,           \
     index,              \
     overlap,            \
     phi_psi

#--------------------------------------------------------------------
from selecting import \
     deselect,           \
     indicate,           \
     select             

#--------------------------------------------------------------------
from exporting import \
     png,                \
     export_coords,      \
     multisave,          \
     save               

#--------------------------------------------------------------------
import editing
from editing import \
     alter,              \
     alter_state,        \
     attach,             \
     bond,               \
     cycle_valence,      \
     deprotect,          \
     edit,               \
     flag,               \
     fuse,               \
     h_add,              \
     h_fill,             \
     invert,             \
     iterate,            \
     iterate_state,      \
     map_set_border,     \
     protect,            \
     push_undo,          \
     redo,               \
     remove,             \
     remove_picked,      \
     rename,             \
     replace,            \
     rotate,             \
     sculpt_purge,       \
     sculpt_deactivate,  \
     sculpt_activate,    \
     sculpt_iterate,     \
     set_dihedral,       \
     set_geometry,       \
     set_title,          \
     smooth,             \
     sort,               \
     torsion,            \
     transform_object,   \
     translate,          \
     translate_atom,     \
     unbond,             \
     undo,               \
     unpick,             \
     update

#--------------------------------------------------------------------

from externing import \
     cd,                 \
     ls,                 \
     paste,              \
     pwd,                \
     system

#--------------------------------------------------------------------
from wizarding import \
     get_wizard,         \
     refresh_wizard,     \
     set_wizard,         \
     wizard
     
#--------------------------------------------------------------------
from fitting import \
     align,             \
     fit,               \
     rms,               \
     rms_cur,           \
     intra_fit,         \
     intra_rms,         \
     intra_rms_cur,     \
     pair_fit          

#--------------------------------------------------------------------
from moving import \
     mset,              \
     mclear,            \
     mdo,               \
     mappend,           \
     mmatrix,           \
     mpng,              \
     forward,           \
     backward,          \
     rewind,            \
     middle,            \
     ending,            \
     mplay,             \
     mstop,             \
     mpng,              \
     mray,              \
     frame,             \
     get_state,         \
     get_frame         

#--------------------------------------------------------------------
import viewing
from viewing import \
     bg_color,           \
     cartoon,            \
     clip,               \
     color,              \
     dirty,              \
     disable,            \
     enable,             \
     full_screen,        \
     get_view,           \
     hide,               \
     label,              \
     load_png,           \
     meter_reset,        \
     move,               \
     orient,             \
     origin,             \
     ray,                \
     rebuild,            \
     recolor,            \
     refresh,            \
     reset,              \
     rock,               \
     set_color,          \
     set_view,           \
     show,               \
     stereo,             \
     turn,               \
     view,               \
     viewport,           \
     zoom


#--------------------------------------------------------------------
import setting
from setting import \
     set,                 \
     get_setting_legacy,  \
     get_setting_tuple,   \
     get_setting_updates, \
     get_setting_text

#--------------------------------------------------------------------
import helping
from helping import \
     show_help,           \
     help,                \
     commands

#--------------------------------------------------------------------
from experimenting import \
     check,              \
     dump,               \
     expfit,             \
     get_bond_print,     \
     fast_minimize,      \
     focus,              \
     import_coords,      \
     load_coords,        \
     mem,                \
     minimize,           \
     spheroid,           \
     test

#--------------------------------------------------------------------
# Modules which contain programs used explicity as "module.xxx"

import util
import movie

# This is the main dictionary

keyword = {

   # keyword : [ command, # min_arg, max_arg, separator, mode ]

   # NOTE: min_arg, max_arg, and separator, are hold-overs from the
   #       original PyMOL parser which will eventually be removed.
   #       all new commands should use NO_CHECK or STRICT modes
   #       which make much better use of built-in python features.
   'abort'         : [ dummy             , 0 , 0 , ''  , parsing.ABORT ],
   'alias'         : [ alias             , 0 , 0 , ''  , parsing.LITERAL1 ],
   'align'         : [ align             , 0 , 0 , ''  , parsing.STRICT ],
   'alter'         : [ alter             , 0 , 0 , ''  , parsing.LITERAL1 ],
   'alter_state'   : [ alter_state       , 0 , 0 , ''  , parsing.LITERAL2 ],
   'attach'        : [ attach            , 0 , 0 , ''  , parsing.LITERAL2 ],   
   'backward'      : [ backward          , 0 , 0 , ''  , parsing.STRICT ],
   'bg_color'      : [ bg_color          , 0 , 0 , ''  , parsing.STRICT ],
   'bond'          : [ bond              , 0 , 0 , ''  , parsing.STRICT ],
   'button'        : [ button            , 0 , 0 , ''  , parsing.STRICT ],
   'cartoon'       : [ cartoon           , 0 , 0 , ''  , parsing.STRICT ],
   'cd'            : [ cd                , 0 , 0 , ''  , parsing.STRICT ],  
   'check'         : [ check             , 0 , 0 , ''  , parsing.STRICT ],
   'clip'          : [ clip              , 0 , 0 , ''  , parsing.STRICT ],
   'cls'           : [ cls               , 0 , 0 , ''  , parsing.STRICT ],
   'color'         : [ color             , 0 , 0 , ''  , parsing.STRICT ],
   'commands'      : [ helping.commands  , 0 , 0 , ''  , parsing.STRICT ],
   'config_mouse'  : [ config_mouse      , 0 , 0 , ''  , parsing.STRICT ],
   'copy'          : [ copy              , 0 , 0 , ''  , parsing.LEGACY ],
   'count_atoms'   : [ count_atoms       , 0 , 0 , ''  , parsing.STRICT ],
   'count_frames'  : [ count_frames      , 0 , 0 , ''  , parsing.STRICT ],   
   'count_states'  : [ count_states      , 0 , 0 , ''  , parsing.STRICT ],
   'cycle_valence' : [ cycle_valence     , 0 , 0 , ''  , parsing.STRICT ],
   'create'        : [ create            , 0 , 0 , ''  , parsing.LEGACY ],   
   'delete'        : [ delete            , 0 , 0 , ''  , parsing.STRICT ],
   'deprotect'     : [ deprotect         , 0 , 0 , ''  , parsing.STRICT ],
   'deselect'      : [ deselect          , 0 , 0 , ''  , parsing.STRICT ],
   'dir'           : [ ls                , 0 , 0 , ''  , parsing.STRICT ],  
   'disable'       : [ disable           , 0 , 0 , ''  , parsing.STRICT ],
   'distance'      : [ distance          , 0 , 0 , ''  , parsing.LEGACY ],   
   'dump'          : [ dump              , 0 , 0 , ''  , parsing.STRICT ],
   'dummy'         : [ dummy             , 0 , 0 , ''  , parsing.STRICT ],   
   'edit'          : [ edit              , 0 , 0 , ''  , parsing.STRICT ],
   'edit_mode'     : [ edit_mode         , 0 , 0 , ''  , parsing.STRICT ],
   'enable'        : [ enable            , 0 , 0 , ''  , parsing.STRICT ],
   'ending'        : [ ending            , 0 , 0 , ''  , parsing.STRICT ],
   'export_dots'   : [ export_dots       , 0 , 0 , ''  , parsing.STRICT ],
   'extend'        : [ extend            , 0 , 0 , ''  , parsing.STRICT ],
   'fast_minimize' : [ fast_minimize     , 1,  4 , ',' , parsing.SIMPLE ], # TO REMOVE
   'feedback'      : [ feedback          , 0,  0 , ''  , parsing.STRICT ],
   'fit'           : [ fit               , 0 , 0 , ''  , parsing.STRICT ],
   'flag'          : [ flag              , 0 , 0 , ''  , parsing.LEGACY ],
   'fork'          : [ helping.spawn    , 1 , 2 , ',' , parsing.SPAWN  ],
   'forward'       : [ forward           , 0 , 0 , ''  , parsing.STRICT ],
   'fragment'      : [ fragment          , 0 , 0 , ''  , parsing.STRICT ],
   'full_screen'   : [ full_screen       , 0 , 0 , ''  , parsing.STRICT ],
   'fuse'          : [ fuse              , 0 , 0 , ''  , parsing.STRICT ],
   'frame'         : [ frame             , 0 , 0 , ''  , parsing.STRICT ],
   'get_area'      : [ get_area          , 0 , 0 , ''  , parsing.STRICT ],
   'get_dihedral'  : [ get_dihedral      , 0 , 0 , ''  , parsing.STRICT ],
   'get_extent'    : [ get_extent        , 0 , 0 , ''  , parsing.STRICT ],   
   'get_position'  : [ get_position      , 0 , 0 , ''  , parsing.STRICT ],
   'get_title'     : [ get_title         , 0 , 0 , ''  , parsing.STRICT ],   
   'get_type'      : [ get_type          , 0 , 0 , ''  , parsing.STRICT ],
   'get_view'      : [ get_view          , 0 , 0 , ''  , parsing.STRICT ],
   'h_add'         : [ h_add             , 0 , 0 , ''  , parsing.STRICT ],
   'h_fill'        : [ h_fill            , 0 , 0 , ''  , parsing.STRICT ],
   'help'          : [ help              , 0 , 0 , ''  , parsing.STRICT ],
   'hide'          : [ hide              , 0 , 0 , ''  , parsing.STRICT ],
   'id_atom'       : [ id_atom           , 0 , 0 , ''  , parsing.STRICT ],
   'identify'      : [ identify          , 0 , 0 , ''  , parsing.STRICT ],
   'index'         : [ index             , 0 , 0 , ''  , parsing.STRICT ],
   'indicate'      : [ indicate          , 0 , 0 , ''  , parsing.STRICT ],   
   'intra_fit'     : [ intra_fit         , 0 , 0 , ''  , parsing.STRICT ],
   'intra_rms'     : [ intra_rms         , 0 , 0 , ''  , parsing.STRICT ],
   'intra_rms_cur' : [ intra_rms_cur     , 0 , 0 , ''  , parsing.STRICT ],
   'invert'        : [ invert            , 0 , 0 , ''  , parsing.STRICT ],
   'isodot'        : [ isodot            , 0 , 0 , ''  , parsing.LEGACY ],   
   'isomesh'       : [ isomesh           , 0 , 0 , ''  , parsing.LEGACY ],
   'isosurface'    : [ isosurface        , 0 , 0 , ''  , parsing.LEGACY ],   
   'iterate'       : [ iterate           , 0 , 0 , ''  , parsing.LITERAL1 ],
   'iterate_state' : [ iterate_state     , 0 , 0 , ''  , parsing.LITERAL2 ],
   'label'         : [ label             , 0 , 0 , ''  , parsing.LITERAL1 ],
   'load'          : [ load              , 0 , 0 , ''  , parsing.STRICT ],
   'load_png'      : [ load_png          , 0 , 0 , ''  , parsing.STRICT ],
   'log'           : [ log               , 0 , 0 , ''  , parsing.STRICT ],
   'log_close'     : [ log_close         , 0 , 0 , ''  , parsing.STRICT ],
   'log_open'      : [ log_open          , 0 , 0 , ''  , parsing.STRICT ],
   'ls'            : [ ls                , 0 , 0 , ''  , parsing.STRICT ],  
   'mask'          : [ mask              , 0 , 0 , ''  , parsing.STRICT ],
   'map_set_border': [ map_set_border    , 0 , 0 , ''  , parsing.STRICT ],
   'map_new'       : [ map_new           , 0 , 0 , ''  , parsing.STRICT ],    
   'mappend'       : [ mappend           , 2 , 2 , ':' , parsing.SINGLE ], 
   'mem'           : [ mem               , 0 , 0 , ''  , parsing.STRICT ],
   'meter_reset'   : [ meter_reset       , 0 , 0 , ''  , parsing.STRICT ],
   'move'          : [ move              , 0 , 0 , ''  , parsing.STRICT ],
   'mset'          : [ mset              , 0 , 0 , ''  , parsing.STRICT ],
   'mdo'           : [ mdo               , 2 , 2 , ':' , parsing.SINGLE ],
   'mpng'          : [ mpng              , 0 , 0 , ''  , parsing.STRICT ],
   'mplay'         : [ mplay             , 0 , 0 , ''  , parsing.STRICT ],
   'mray'          : [ mray              , 0 , 0 , ''  , parsing.STRICT ],
   'mstop'         : [ mstop             , 0 , 0 , ''  , parsing.STRICT ],
   'mclear'        : [ mclear            , 0 , 0 , ''  , parsing.STRICT ],
   'middle'        : [ middle            , 0 , 0 , ''  , parsing.STRICT ],
   'minimize'      : [ minimize          , 0 , 4 , ',' , parsing.SIMPLE ], # TO REMOVE
   'mmatrix'       : [ mmatrix           , 0 , 0 , ''  , parsing.STRICT ],
   'mouse'         : [ mouse             , 0 , 0 , ''  , parsing.STRICT ],
   'multisave'     : [ multisave         , 0 , 0 , ''  , parsing.STRICT ],   
   'origin'        : [ origin            , 0 , 0 , ''  , parsing.STRICT ],
   'orient'        : [ orient            , 0 , 0 , ''  , parsing.STRICT ],
   'overlap'       : [ overlap           , 0 , 0 , ''  , parsing.STRICT ],
   'pair_fit'      : [ pair_fit          , 2 ,98 , ',' , parsing.SIMPLE ],
   'phi_psi'       : [ phi_psi           , 0 , 0 , ''  , parsing.STRICT ],
   'protect'       : [ protect           , 0 , 0 , ''  , parsing.STRICT ],
   'push_undo'     : [ push_undo         , 0 , 0 , ''  , parsing.STRICT ],   
   'pwd'           : [ pwd               , 0 , 0 , ''  , parsing.STRICT ],
   'ray'           : [ ray               , 0 , 0 , ''  , parsing.STRICT ],
   'rebuild'       : [ rebuild           , 0 , 0 , ''  , parsing.STRICT ],
   'recolor'       : [ recolor           , 0 , 0 , ''  , parsing.STRICT ],   
   'redo'          : [ redo              , 0 , 0 , ''  , parsing.STRICT ],
   'refresh'       : [ refresh           , 0 , 0 , ''  , parsing.STRICT ],
   'remove'        : [ remove            , 0 , 0 , ''  , parsing.STRICT ],
   'remove_picked' : [ remove_picked     , 0 , 0 , ''  , parsing.STRICT ],
   'rename'        : [ rename            , 0 , 0 , ''  , parsing.STRICT ],
   'replace'       : [ replace           , 0 , 0 , ''  , parsing.STRICT ],
   'reset'         : [ reset             , 0 , 0 , ''  , parsing.STRICT ],
   'resume'        : [ resume            , 0 , 0 , ''  , parsing.STRICT ],
   'rewind'        : [ rewind            , 0 , 0 , ''  , parsing.STRICT ],
   'rock'          : [ rock              , 0 , 0 , ''  , parsing.STRICT ],
   'rotate'        : [ rotate            , 0 , 0 , ''  , parsing.STRICT ],   
   'run'           : [ helping.run       , 1 , 2 , ',' , parsing.RUN    ],
   'rms'           : [ rms               , 0 , 0 , ''  , parsing.STRICT ],
   'rms_cur'       : [ rms_cur           , 0 , 0 , ''  , parsing.STRICT ],
   'save'          : [ save              , 0 , 0 , ''  , parsing.STRICT ],
   'sculpt_purge'  : [ sculpt_purge      , 0 , 0 , ''  , parsing.STRICT ],   
   'sculpt_deactivate': [ sculpt_deactivate , 0 , 0 , ''  , parsing.STRICT ],
   'sculpt_activate': [ sculpt_activate  , 0 , 0 , ''  , parsing.STRICT ],
   'sculpt_iterate': [ sculpt_iterate    , 0 , 0 , ''  , parsing.STRICT ],
   'select'        : [ select            , 0 , 0 , ''  , parsing.LEGACY ],
   'set'           : [ set               , 0 , 0 , ''  , parsing.LEGACY ],
   'set_color'     : [ set_color         , 0 , 0 , ''  , parsing.LEGACY ],
   'set_dihedral'  : [ set_dihedral      , 0 , 0 , ''  , parsing.STRICT ],   
   'set_geometry'  : [ set_geometry      , 0 , 0 , ''  , parsing.STRICT ],      
   'set_title'     : [ set_title         , 0 , 0 , ''  , parsing.STRICT ],   
   'set_key'       : [ set_key           , 0 , 0 , ''  , parsing.STRICT ], # API only
   'set_view'      : [ set_view          , 0 , 0 , ''  , parsing.STRICT ],   
   'show'          : [ show              , 0 , 0 , ''  , parsing.STRICT ],
   'smooth'        : [ smooth            , 0 , 0 , ''  , parsing.STRICT ],
   'sort'          : [ sort              , 0 , 0 , ''  , parsing.STRICT ],
   'spawn'         : [ helping.spawn     , 1 , 2 , ',' , parsing.SPAWN  ],
   'spheroid'      : [ spheroid          , 0 , 0 , ''  , parsing.STRICT ],
   'splash'        : [ splash            , 0 , 0 , ''  , parsing.STRICT ],
   '_special'      : [ _special          , 0 , 0 , ''  , parsing.STRICT ],
   'stereo'        : [ stereo            , 0 , 0 , ''  , parsing.STRICT ],
   'symexp'        : [ symexp            , 0 , 0 , ''  , parsing.LEGACY ],
   'system'        : [ system            , 0 , 0 , ''  , parsing.LITERAL ],
   'test'          : [ test              , 0 , 0 , ''  , parsing.STRICT ],
   'torsion'       : [ torsion           , 0 , 0 , ''  , parsing.STRICT ],
   'translate'     : [ translate         , 0 , 0 , ''  , parsing.STRICT ],
   'turn'          : [ turn              , 0 , 0 , ''  , parsing.STRICT ],
   'quit'          : [ quit              , 0 , 0 , ''  , parsing.STRICT ],
   '_quit'         : [ _quit             , 0 , 0 , ''  , parsing.STRICT ],
   'png'           : [ png               , 0 , 0 , ''  , parsing.STRICT ],
   'unbond'        : [ unbond            , 0 , 0 , ''  , parsing.STRICT ],
   'unpick'        : [ unpick            , 0 , 0 , ''  , parsing.STRICT ],
   'undo'          : [ undo              , 0 , 0 , ''  , parsing.STRICT ],
   'unmask'        : [ unmask            , 0 , 0 , ''  , parsing.STRICT ],
   'unprotect'     : [ deprotect         , 0 , 0 , ''  , parsing.STRICT ],
   'update'        : [ update            , 0 , 0 , ''  , parsing.STRICT ],
   'view'          : [ view              , 0 , 0 , ''  , parsing.STRICT ],   
   'viewport'      : [ viewport          , 0 , 0 , ''  , parsing.STRICT ],
   'wizard'        : [ wizard            , 0 , 0 , ''  , parsing.STRICT ],
   'zoom'          : [ zoom              , 0 , 0 , ''  , parsing.STRICT ],
# utility programs
   'util.cbag'     : [ util.cbag         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbac'     : [ util.cbac         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbay'     : [ util.cbay         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbas'     : [ util.cbas         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbap'     : [ util.cbap         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbak'     : [ util.cbak         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbaw'     : [ util.cbaw         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbab'     : [ util.cbab         , 0 , 0 , ''  , parsing.STRICT ],
   'util.cbc'      : [ util.cbc          , 0 , 0 , ''  , parsing.STRICT ],
   'util.mrock'    : [ util.mrock        , 0 , 0 , ''  , parsing.STRICT ], # LEGACY
   'util.mroll'    : [ util.mroll        , 0 , 0 , ''  , parsing.STRICT ], # LEGACY
   'util.ss'       : [ util.ss           , 0 , 0 , ''  , parsing.STRICT ],# secondary structure
   'util.rainbow'  : [ util.rainbow      , 0 , 0 , ''  , parsing.STRICT ],# secondary structure
# movie programs
   'movie.rock'    : [ movie.rock        , 0 , 0 , ''  , parsing.STRICT ],
   'movie.roll'    : [ movie.roll        , 0 , 0 , ''  , parsing.STRICT ],
   'movie.load'    : [ movie.load        , 0 , 0 , ''  , parsing.STRICT ],
   'movie.zoom'    : [ movie.zoom        , 0 , 0 , ''  , parsing.STRICT ],
   'movie.screw'   : [ movie.screw       , 0 , 0 , ''  , parsing.STRICT ],
   'movie.nutate'  : [ movie.nutate      , 0 , 0 , ''  , parsing.STRICT ],
   'movie.tdroll'  : [ movie.tdroll      , 0 , 0 , ''  , parsing.STRICT ],
   }

kwhash = Shortcut(keyword.keys())

# informational, or API-only functions which don't exist in the
# PyMOL command language namespace

help_only = {  
   'api'           : [ helping.api          , 0 , 0 , '' , 0 ],
   'selections'    : [ helping.selections   , 0 , 0 , '' , 0 ],
   'keyboard'      : [ helping.keyboard     , 0 , 0 , '' , 0 ],
   'mouse'         : [ helping.mouse        , 0 , 0 , '' , 0 ],
   'examples'      : [ helping.examples     , 0 , 0 , '' , 0 ],
   'read_molstr'   : [ read_molstr           , 0 , 0 , '' , 0 ],
   'release'       : [ helping.release      , 0 , 0 , '' , 0 ],   
   'launching'     : [ helping.launching    , 0 , 0 , '' , 0 ],
   'load_model'    : [ load_model            , 0 , 0 , '' , 0 ],
   'movies'        : [ helping.movies       , 0 , 0 , '' , 0 ],
   'editing'       : [ helping.editing      , 0 , 0 , '' , 0 ],  
   'edit_keys'     : [ helping.edit_keys    , 0 , 0 , '' , 0 ],
   'povray'        : [ helping.povray       , 0 , 0 , '' , 0 ],
   'get_names'     : [ get_names             , 0 , 0 , '' , 0 ],
   'get_type'      : [ get_type              , 0 , 0 , '' , 0 ],
   'faster'        : [ helping.faster       , 0 , 0 , '' , 0 ],
   'sync'          : [ sync                 , 0 , 0 , '' , 0],
   'transparency'  : [ helping.transparency , 0 , 0 , '' , 0 ],  
   '@'             : [ helping.at_sign      , 0 , 0 , '' , 0 ],  
}

help_sc = Shortcut(keyword.keys()+help_only.keys())


special = {
   1        : [ 'F1'        , None                   , () , {} ],
   2        : [ 'F2'        , None                   , () , {} ],
   3        : [ 'F3'        , None                   , () , {} ],
   4        : [ 'F4'        , None                   , () , {} ],
   5        : [ 'F5'        , None                   , () , {} ],
   6        : [ 'F6'        , None                   , () , {} ],
   7        : [ 'F7'        , None                   , () , {} ],
   8        : [ 'F8'        , None                   , () , {} ],
   9        : [ 'F9'        , None                   , () , {} ],
   10       : [ 'F10'       , None                   , () , {} ],
   11       : [ 'F11'       , redo                   , () , {} ],
   12       : [ 'F12'       , undo                   , () , {} ],
   100      : [ 'left'      , backward               , () , {} ],
   101      : [ 'up'        , None                   , () , {} ],
   102      : [ 'right'     , forward                , () , {} ],
   103      : [ 'down'      , None                   , () , {} ],
   104      : [ 'pgup'      , rewind                 , () , {} ],
   105      : [ 'pgdn'      , ending                 , () , {} ],
   106      : [ 'home'      , zoom                   , () , {} ],
   107      : [ 'end'       , ending                 , () , {} ],
   108      : [ 'insert'    , rock                   , () , {} ]   
}

ctrl = {
   'A' : [ redo                   , () , {}],
   'B' : [ replace                , ('Br',1,1), {} ],
   'C' : [ replace                , ('C',4,4), {} ],
   'D' : [ remove_picked          , () , {} ],
   'E' : [ invert                 , () , {} ],   
   'F' : [ replace                , ('F',1,1), {} ],   
   'G' : [ replace                , ('H',1,1), {} ],
   'I' : [ replace                , ('I',1,1), {} ],
   'J' : [ alter                  , ('pk1','formal_charge=-1.0'), {} ],
   'K' : [ alter                  , ('pk1','formal_charge =1.0'), {} ],
   'L' : [ replace                , ('Cl',1,1) , {}],
   'N' : [ replace                , ('N',4,3) , {}],
   'O' : [ replace                , ('O',4,2) , {}],   
   'P' : [ replace                , ('P',4,1) , {}],
   'Q' : [ dist                   , () , {}],   
   'R' : [ h_fill                 , () , {} ],   
   'S' : [ replace                , ('S',4,2) , {}],
   'T' : [ bond                   , () , {} ],   
   'U' : [ alter                  , ('pk1','formal_charge =0.0') , {}],
   'W' : [ cycle_valence          , () , {}],   
   'X' : [ remove                 , ('pkfrag1',) , {} ],
   'Y' : [ attach                 , ('H',1,1) , {} ],
   'Z' : [ undo                   , () , {} ],   
   }

alt = {
   '1' : [ editor.attach_fragment  , ("pk1","formamide",5,0), {}],
   '2' : [ editor.attach_fragment  , ("pk1","formamide",4,0), {}],
   '3' : [ editor.attach_fragment  , ("pk1","sulfone",3,1), {}],
   '4' : [ editor.attach_fragment  , ("pk1","cyclobutane",4,0), {}],
   '5' : [ editor.attach_fragment  , ("pk1","cyclopentane",5,0), {}],
   '6' : [ editor.attach_fragment  , ("pk1","cyclohexane",7,0), {}],
   '7' : [ editor.attach_fragment  , ("pk1","cycloheptane",8,0), {}],
   '8' : [ editor.attach_fragment  , ("pk1","cyclopentadiene",5,0), {}],
   '9' : [ editor.attach_fragment  , ("pk1","benzene",6,0), {}],
   '0' : [ editor.attach_fragment  , ("pk1","formaldehyde",2,0), {}],
   'a' : [ editor.attach_amino_acid, ("pk1","ala"), {}],
   'b' : [ editor.attach_amino_acid, ("pk1","ace"), {}],                                 
   'c' : [ editor.attach_amino_acid, ("pk1","cys"), {}],
   'd' : [ editor.attach_amino_acid, ("pk1","asp"), {}],
   'e' : [ editor.attach_amino_acid, ("pk1","glu"), {}],
   'f' : [ editor.attach_amino_acid, ("pk1","phe"), {}],
   
   'g' : [ editor.attach_amino_acid, ("pk1","gly"), {}],
   'h' : [ editor.attach_amino_acid, ("pk1","his"), {}],
   'i' : [ editor.attach_amino_acid, ("pk1","ile"), {}],

   'j' : [ editor.attach_fragment,   ("pk1","acetylene",2,0), {}],
   'k' : [ editor.attach_amino_acid, ("pk1","lys"), {}],
   'l' : [ editor.attach_amino_acid, ("pk1","leu"), {}],
   
   'm' : [ editor.attach_amino_acid, ("pk1","met"), {}],
   'n' : [ editor.attach_amino_acid, ("pk1","asn"), {}],
   'p' : [ editor.attach_amino_acid, ("pk1","pro"), {}],
   'q' : [ editor.attach_amino_acid, ("pk1","gln"), {}],
   'r' : [ editor.attach_amino_acid, ("pk1","arg"), {}],
   
   's' : [ editor.attach_amino_acid, ("pk1","ser"), {}],
   't' : [ editor.attach_amino_acid, ("pk1","thr"), {}],
   'v' : [ editor.attach_amino_acid, ("pk1","val"), {}],
   'w' : [ editor.attach_amino_acid, ("pk1","trp"), {}],
   'y' : [ editor.attach_amino_acid, ("pk1","tyr"), {}],
   'z' : [ editor.attach_amino_acid, ("pk1","nme"), {}],                              
   }

selection_sc = lambda sc=Shortcut,gn=get_names:sc(gn('public')+['all'])
object_sc = lambda sc=Shortcut,gn=get_names:sc(gn('objects'))
map_sc = lambda sc=Shortcut,gnot=get_names_of_type:sc(gnot('object:map'))

# Table for argument autocompletion

auto_arg =[
   {
   'color'          : [ _get_color_sc          , 'color'           , ', ' ],
   'cartoon'        : [ viewing.cartoon_sc     , 'cartoon'         , ', ' ],      
   'set'            : [ setting.setting_sc     , 'setting'         , '='  ],
   'flag'           : [ editing.flag_sc        , 'flag'            , ', ' ],
   'show'           : [ repres_sc              , 'representation'  , ', ' ],
   'hide'           : [ repres_sc              , 'representation'  , ', ' ],
   'stereo'         : [ stereo_sc              , 'option'          , ''   ],
   'full_screen'    : [ toggle_sc              , 'option'          , ''   ],
   'clip'           : [ viewing.clip_action_sc , 'clipping action' , ', ' ],
   'feedback'       : [ fb_action_sc           , 'action'          , ', ' ],
   'button'         : [ controlling.button_sc  , 'button'          , ', ' ],
   'align'          : [ selection_sc           , 'selection'       , ','  ],
   'zoom'           : [ selection_sc           , 'selection'       , ''   ],
   'origin'         : [ selection_sc           , 'selection'       , ''   ],
   'protect'        : [ selection_sc           , 'selection'       , ''   ],
   'deprotect'      : [ selection_sc           , 'selection'       , ''   ],   
   'mask'           : [ selection_sc           , 'selection'       , ''   ],
   'unmask'         : [ selection_sc           , 'selection'       , ''   ],
   'delete'         : [ selection_sc           , 'selection'       , ''   ],
   'alter'          : [ selection_sc           , 'selection'       , ''   ],
   'iterate'        : [ selection_sc           , 'selection'       , ''   ],
   'iterate_state'  : [ selection_sc           , 'selection'       , ''   ],
   'help'           : [ help_sc                , 'selection'       , ''   ],         
   },
   {
   'align'          : [ selection_sc           , 'selection'       , ''   ],
   'feedback'       : [ fb_module_sc           , 'module'          , ', ' ],
   'button'         : [ controlling.but_mod_sc , 'modifier'        , ', ' ],
   'show'           : [ selection_sc           , 'selection'       , ''   ],
   'hide'           : [ selection_sc           , 'selection'       , ''   ],
   'color'          : [ selection_sc           , 'selection'       , ''   ],
   'select'         : [ selection_sc           , 'selection'       , ''   ],
   'save'           : [ selection_sc           , 'selection'       , ', ' ],
   'flag'           : [ selection_sc           , 'selection'       , ', ' ],   
   'load'           : [ selection_sc           , 'selection'       , ', ' ],
   'create'         : [ selection_sc           , 'selection'       , ', ' ],
   'symexp'         : [ object_sc              , 'object'          , ', ' ],   
   'isomesh'        : [ map_sc                 , 'map object'      , ', ' ],
   'view'           : [ viewing.view_sc        , 'view action'     , ''   ],            
   },
   {
   'feedback'       : [ fb_mask_sc             , 'mask'            , ''   ],
   'button'         : [ controlling.but_act_sc , 'button action'   , ''   ],
   'flag'           : [ editing.flag_action_sc , 'flag action'     , ''   ],      
   }
   ]
   
color_sc = None




from chempy import Storage,Atom
from chempy.models import Indexed
import string

class PDB(Storage):
   
#---------------------------------------------------------------------------------
   def fromList(self,list):   # currently no handling of conect records 

      model = Indexed()
      
# read atoms
      cnt = 0
      at = None
      for rec in list:
         if (rec[0:4]=='ATOM') or (rec[0:6]=='HETATM'):
            at = Atom()
            if rec[0]=='A': at.hetatm=0  # default is 1
            at.index = cnt
            at.name = string.strip(rec[12:16])
            at.alt = string.strip(rec[16:17])
            at.resn = string.strip(rec[17:20])
            at.chain = string.strip(rec[21:22])
            at.resi = string.strip(rec[22:27]) # note: insertion is part of resi
            at.resi_number = int(rec[22:26])
            at.coord = [float(rec[30:38]), 
                        float(rec[38:46]),
                        float(rec[46:54])]
            try:
               at.q = float(rec[54:60])
            except ValueError:
               at.q = 1.0
            try:               
               at.b = float(rec[60:66])
            except ValueError:
               at.b = 0.0
            at.segi = string.strip(rec[72:76])
            at.symbol = string.strip(rec[76:78])
            if not len(at.symbol):
               at.symbol = at.name[0:1]
               if at.symbol in '012345678':
                  at.symbol = at.name[1:2]
            cnt = cnt + 1
            model.add_atom(at)
         elif (rec[0:3]=='TER'):
            if at:
               at.ter=1
      return(model)

#---------------------------------------------------------------------------------
   def toList(self,model):

      list = []

      cnt = 1
      for at in model.atom:
         if at.hetatm:
            het = 'HETATM'
         else:
            het = 'ATOM  '
         latn = len(at.name)
         if latn and (latn<4):
            if at.name[0] in '0123456789':
               name=at.name
            else:
               name=' '+at.name
         else:
            name=at.name
         lrsi = len(at.resi)
         if lrsi and (lrsi<5):
            if at.resi[lrsi-1] in '0123456789':
               resi=at.resi+' '
            else:
               resi=at.resi
         list.append(
            "%6s%5i %-4s%1s%3s %1s%5s   %8.3f%8.3f%8.3f%6.2f%6.2f      %-4s%2s\n" % 
            (het,cnt,name,at.alt,at.resn,at.chain,resi,
            at.coord[0],at.coord[1],at.coord[2],at.q,at.b,at.segi,at.symbol))
         if hasattr(at,'ter'):
            if at.ter:
               cnt = cnt + 1
               list.append(
                  "%6s%5i      %3s %1s%4s\n" %
                  ('TER   ',cnt,at.resn,at.chain,at.resi))
         cnt = cnt + 1
      list.append("END\n")
      return(list)




from chempy.models import Indexed,Connected
from chempy import Storage,Atom,Bond

# ad-hoc maestro file parser

import re
import string
import copy

strip_re = re.compile(r'\#.*\#')
token_re = re.compile(r'"([^"]*)"|([^ ]+)')
array_re = re.compile(r'(.*)\[([0-9]+)\]')

coerce = {
   's' : str,
   'i' : int,
   'r' : float }

class MAEParser:

   def __init__(self,lst=None):
      self.i = 0 # index in list
      self.t = [] # token list 
      self.d = [] # hiearchy of data read 
      self.lst = lst
      self.lst_len = len(lst)

   def nxt_lin(self):
      if self.lst:
         if self.i<self.lst_len:
            self.i = self.i + 1
            return self.lst[self.i-1]
      return None

   def nxt_tok(self):
      while 1:
         if len(self.t):
            return self.t.pop(0)
         else:
            l = self.nxt_lin()
            if not l:
               return None
            l = string.strip(strip_re.sub('',l))
            self.t = token_re.findall(l)
            self.t = map(lambda x:string.join(x,''),self.t)
      return None

   def push_tok(self,tok):
      self.t.insert(0,tok)
      
   def parse_top(self):
      dct = {}
      stk = [] # keyword stack
      mode = 0 # 0 = definition, 1 = data
      while 1:
         tok = self.nxt_tok()
         if tok==None:
            break
         if tok==':::':
            mode = 1
         if not len(stk):
            mode = 0
         if mode:
            lab = stk.pop(0)
            dct[lab] = apply(coerce[lab[0]],(tok,))
         else:
            stk.append(tok)
         if tok=='}':
            break
      return dct

   def parse_array(self,n_rec): # creates a list of homogenous lists
                                # each containing data for one field
      dct = {}
      data = [] # actual array data
      stk = [] # keyword stack
      coer = [] # coersion functions
      n_fld = 0
      mode = 0 # 0 = definition, 1 = data
      cc = 0
      while 1:
         tok = self.nxt_tok()
         if tok==None:
            break
         if tok=='}':
            break
         if tok==':::':
            if not mode:
               mode = 1
               n_fld = len(stk)
               c = 0
               for a in stk: # create row index for the array
                  dct[a] = c
                  data.append([]) # add row for each field
                  coer.append(coerce[a[0]])
                  c = c + 1
         elif not mode:
            stk.append(tok)
         else: # here we actually read the array
            self.push_tok(tok)
            for cc in range(n_rec):
               tok = self.nxt_tok() # chuck index
               if tok==':::': # truncated/incomplete array
                  break
               c = 0
               for c in range(n_fld):
                  tok = self.nxt_tok()
                  if tok==None:
                     break
                  data[c].append(apply(coer[c],(tok,)))
                  if tok=='}':
                     break
      return (n_rec,dct,data) # return a tuple
      
   def parse_m_ct(self):
      dct = {}
      stk = [] # keyword stack
      mode = 0 # 0 = definition, 1 = data
      while 1:
         tok = self.nxt_tok()
         if tok==None:
            break
         if tok==':::':
            mode = 1
         elif mode:
            if not len(stk):
               mode = 0
               self.push_tok(tok) # go around
            else:
               dct[stk.pop(0)] = tok
         else:
            arm = array_re.findall(tok)
            if len(arm):
               arm = arm[0]
               n_rec=int(arm[1])
               if arm[0] in ['m_atom','m_bond']:
                  self.nxt_tok() # skip '{'
                  dct[arm[0]]=self.parse_array(n_rec)
            else:
               stk.append(tok)
         if tok=='}':
            break
      return dct
      
   def parse(self):
      while 1:
         tok = self.nxt_tok()
         if tok==None:
            break
         if tok=='{':
            self.d.append('top',self.parse_top())
         elif tok in ['f_m_ct','p_m_ct']:
            self.nxt_tok() # skip '{'
            self.d.append(tok,self.parse_m_ct())
      return self.d
   
class MAE(Storage):

   def _read_m_atom(self,m_atom,model):
      ma = model.atom
      at_ent = m_atom[1]
      at_dat = m_atom[2]
      
      nAtom = m_atom[0]      
      if at_ent.has_key('i_m_mmod_type'):
         a1 = at_dat[at_ent['i_m_mmod_type']]
         
         for b in range(nAtom):
            nt = a1[b]
            ma[b].numeric_type = nt 
            ma[b].symbol = MMOD_atom_data[nt][1]
            ma[b].text_type = MMOD_atom_data[nt][0]
            
      if at_ent.has_key('r_m_x_coord') and \
         at_ent.has_key('r_m_y_coord') and \
         at_ent.has_key('r_m_z_coord'):
         a1 = at_dat[at_ent['r_m_x_coord']]
         a2 = at_dat[at_ent['r_m_y_coord']]
         a3 = at_dat[at_ent['r_m_z_coord']]
         for b in range(nAtom):
            ma[b].coord = [a1[b],a2[b],a3[b] ]
            
      if at_ent.has_key('i_m_residue_number'):
         a1 = at_dat[at_ent['i_m_residue_number']]
         for b in range(nAtom):
            resi = a1[b]
            ma[b].resi = str(resi)
            ma[b].resi_number = resi
            
      if at_ent.has_key('s_m_mmod_res'):
         a1 = at_dat[at_ent['s_m_mmod_res']]
         for b in range(nAtom):         
            ma[b].resi_code = a1[b]
            
      if at_ent.has_key('s_m_chain_name'):
         a1 = at_dat[at_ent['s_m_chain_name']]
         for b in range(nAtom):
            ma[b].chain = a1[b]
            
      if at_ent.has_key('i_m_color'):
         a1 = at_dat[at_ent['i_m_color']]
         for b in range(nAtom):
            ma[b].color_code = a1[b]
            
      if at_ent.has_key('r_m_charge1'):
         a1 = at_dat[at_ent['r_m_charge1']]
         for b in range(nAtom):
            ma[b].partial_charge = a1[b]
            
      if at_ent.has_key('s_m_pdb_residue_name'):
         a1 = at_dat[at_ent['s_m_pdb_residue_name']]
         for b in range(nAtom):
            resn = string.strip(a1[b])
            if len(resn):
               ma[b].resn = resn
               
      if at_ent.has_key('i_m_formal_charge'):
         a1 = at_dat[at_ent['i_m_formal_charge']]
         for b in range(nAtom):
            ma[b].formal_charge = a1[b]
            
      if at_ent.has_key('s_m_atom_name'):
         a1 = at_dat[at_ent['s_m_atom_name']]
         for b in range(nAtom):
            nam = string.strip(a1[b])
            if len(nam):
               ma[b].name = nam
            
      if at_ent.has_key('s_m_pdb_atom_name'):
         a1 = at_dat[at_ent['s_m_pdb_atom_name']]
         for b in range(nAtom):
            nam = string.strip(a1[b])
            if len(nam):
               ma[b].name = nam

   def _read_m_bond(self,m_bond,model):
      bd_ent = m_bond[1]
      bd_dat = m_bond[2]
      
      nBond = m_bond[0]      

      if len(bd_dat[0]): # not empty right?
         if bd_ent.has_key('i_m_from') and \
            bd_ent.has_key('i_m_to') and \
            bd_ent.has_key('i_m_order'):
            a1 = bd_dat[bd_ent['i_m_from']]
            a2 = bd_dat[bd_ent['i_m_to']]
            a3 = bd_dat[bd_ent['i_m_order']]
            for b in range(nBond):
               bd1 = a1[b] - 1
               bd2 = a2[b] - 1
               bd3 = a3[b]

               if bd1<bd2:
                  bnd = Bond()
                  bnd.index = [ bd1,bd2 ]
                  bnd.order = bd3
                  model.bond.append(bnd)
                  
#---------------------------------------------------------------------------------
   def fromList(self,MMODList): # returns a list of indexed models

      mp = MAEParser(lst=MMODList)
      mp_rec = mp.parse()

      full_model = None
      result = []
      
      for mp_ent in mp_rec:
         if mp_ent[0] == 'f_m_ct':
            f_m_ct = mp_ent[1]
            model = Indexed()
            if f_m_ct.has_key('s_m_title'):
               model.molecule.title = string.strip(f_m_ct['s_m_title'])

            if f_m_ct.has_key('m_atom'):
               m_atom = f_m_ct['m_atom']
               nAtom = m_atom[0]
               for a in range(nAtom):
                  model.atom.append(Atom())
               self._read_m_atom(m_atom,model)

            if f_m_ct.has_key('m_bond'):
               m_bond = f_m_ct['m_bond']
               self._read_m_bond(m_bond,model)
            full_model = model
            result.append(model)
            
         elif mp_ent[0]=='p_m_ct' and full_model!=None:
            model = copy.deepcopy(full_model)

            f_m_ct = mp_ent[1]
            if f_m_ct.has_key('s_m_title'):
               model.molecule.title = string.strip(f_m_ct['s_m_title'])

            if f_m_ct.has_key('m_atom'):
               m_atom = f_m_ct['m_atom']
               nAtom = m_atom[0]
               self._read_m_atom(m_atom,model)

            if f_m_ct.has_key('m_bond'):
               m_bond = f_m_ct['m_bond']
               self._read_m_bond(m_bond,model)
            full_model = model
            result.append(model)

      return result

#---------------------------------------------------------------------------------
'#Ntype Atype Elem Hybr Att Chg\n',
MMOD_atom_data = {
   1: ['C1','C' ,'sp' , 2, 0],
   2: ['C2','C' ,'sp2', 3, 0],
   3: ['C3','C' ,'sp3', 4, 0],
   4: ['CA','C' ,'sp3', 3, 0],
   5: ['CB','C' ,'sp3', 2, 0],
   6: ['CC','C' ,'sp3', 1, 0],
   7: ['CD','C' ,'sp2', 2, 0],
   8: ['CE','C' ,'sp2', 1, 0],
   9: ['CF','C' ,'sp' , 1, 0],
  10: ['CM','C' ,'unk',-1,-1],
  11: ['CP','C' ,'unk',-1, 1],
  12: ['CR','C' ,'unk',-1, 0],
  14: ['C0','C' ,'unk',-1, 0],
  15: ['O2','O' ,'sp2', 1, 0],
  16: ['O3','O' ,'sp3', 2, 0],
  17: ['OA','O' ,'sp3', 1, 0],
  18: ['OM','O' ,'sp3', 1,-1],
  19: ['OW','O' ,'sp3', 0, 0],
  20: ['OP','O' ,'sp2', 2, 1],
  21: ['OQ','O' ,'sp3', 3, 1],
  23: ['O0','O' ,'unk',-1, 0],
  24: ['N1','N' ,'sp' , 1, 0],
  25: ['N2','N' ,'sp2', 2, 0],
  26: ['N3','N' ,'sp3', 3, 0],
  27: ['NA','N' ,'sp3', 2, 0],
  28: ['NB','N' ,'sp3', 1, 0],
  29: ['NC','N' ,'sp3', 0, 0],
  30: ['ND','N' ,'sp2', 1, 0],
  31: ['N4','N' ,'sp2', 3, 1],
  32: ['N5','N' ,'sp3', 4, 1],
  33: ['NE','N' ,'sp3', 3, 1],
  34: ['NF','N' ,'sp3', 2, 1],
  35: ['NG','N' ,'sp3', 1, 1],
  36: ['NH','N' ,'sp2', 2, 1],
  37: ['NI','N' ,'sp2', 1, 1],
  40: ['N0','N' ,'unk',-1, 0],
  41: ['H1','H' ,'s'  , 1, 0],
  42: ['H2','H' ,'s'  , 1, 0],
  43: ['H3','H' ,'s'  , 1, 0],
  44: ['H4','H' ,'s'  , 0, 0],
  45: ['H5','H' ,'s'  , 0, 0],
  48: ['H0','H' ,'s'  ,-1, 0],
  49: ['S1','S' ,'sp3', 2, 0],
  50: ['SA','S' ,'sp3', 1, 0],
  51: ['SM','S' ,'sp3', 0,-1],
  52: ['S0','S' ,'unk',-1, 0],
  53: ['P0','P' ,'unk',-1, 0],
  54: ['B2','B' ,'sp2', 2, 0],
  55: ['B3','B' ,'sp3', 3, 0],
  56: ['F0','F' ,'sp3', 1, 0],
  57: ['Cl','Cl','sp3', 1, 0],
  58: ['Br','Br','sp3', 1, 0],
  59: ['I0','I' ,'sp3', 1, 0],
  60: ['Si','Si','unk',-1, 0],
  61: ['Du','Du','unk',-1, 0],
  62: ['Du','Du','unk',-1, 0],
  63: ['Lp','Lp','unk', 1, 0],
  64: ['Du','Du','unk',-1, 0]};





from pymol.wizard import Wizard
from pymol import cmd
from chempy import io

import pymol
import os
import string

sele_name = "_mutate_sel"

obj_name = "mutation"
tmp_name = "_tmp_mut"
tmp_obj1 = "_tmp_obj1"
tmp_obj2 = "_tmp_obj2"
tmp_obj3 = "_tmp_obj3"

default_mode = "current"
default_rep = "lines"

class Mutagenesis(Wizard):

   count = 0
   cutoff = 3.5
   
   def __init__(self):

      Wizard.__init__(self)

      self.library = io.pkl.fromFile(os.environ['PYMOL_PATH']+
                                     "/modules/chempy/sidechains/sc_library.pkl")
      
      self.status = 0 # 0 no selection, 1 mutagenizing
      self.error = None
      self.object_name = None
      self.modes = [
         'current'
         ]
      self.mode = default_mode
      self.rep = default_rep
      residues = self.library.keys()
      residues.extend(['GLY','PRO','ALA'])
      residues.sort()
      self.modes.extend(residues)
      self.mode_name={}
      for a in self.modes:
         self.mode_name[a] = "-> "+a
      self.mode_name['current']="No Mutation"

      smm = []
      smm.append([ 2, 'Substitution', '' ])
      for a in self.modes:
         smm.append([ 1, self.mode_name[a], 'cmd.get_wizard().set_mode("'+a+'")'])
      self.menu['mode']=smm

      self.reps = [
         'lines',
         'sticks',
         'spheres',
         'dots'
         ]

      self.rep_name = {
         'lines' : "Show Lines",
         'sticks' : "Show Sticks",
         'spheres' : "Show Spheres",
         'dots' : "Show Dots",
         }

      smm = []
      smm.append([ 2, 'Representation', '' ])
      for a in self.reps:
         smm.append([ 1, self.rep_name[a], 'cmd.get_wizard().set_rep("'+a+'")'])
      self.menu['rep']=smm

      if 'pk1' in cmd.get_names('selections'):
         cmd.select(sele_name,"(byres pk1)")
         cmd.unpick()
         cmd.enable(sele_name)
         self.status = 1
         self.error = None
         self.do_library()
         cmd.refresh_wizard()
      

   def set_mode(self,mode):
      if mode in self.modes:
         self.mode = mode
      if self.status==1:
         self.do_library()
      cmd.refresh_wizard()
      
   def set_rep(self,rep):
      if rep in self.reps:
         self.rep=rep
      cmd.hide("("+obj_name+")")
      cmd.show('lines',obj_name) # always show lines      
      cmd.show(self.rep,obj_name)
      cmd.refresh_wizard()
      
   def get_panel(self):
      return [
         [ 1, 'Mutagenesis',''],
         [ 3, self.mode_name[self.mode],'mode'],
         [ 3, self.rep_name[self.rep],'rep'],                  
         [ 2, 'Apply' , 'cmd.get_wizard().apply()'],         
         [ 2, 'Clear' , 'cmd.get_wizard().clear()'],
         [ 2, 'Done','cmd.set_wizard()'],
         ]

   def cleanup(self):
      global default_mode,default_rep
      default_mode = self.mode
      default_rep = self.rep
      self.clear()
      
   def clear(self):
      self.status=0
      cmd.delete(tmp_obj1)
      cmd.delete(tmp_obj2)
      cmd.delete(tmp_obj3)
      cmd.delete(sele_name)
      cmd.delete(obj_name)
      cmd.refresh_wizard()
      
   def apply(self):
      if self.status==1:
         # find the name of the object which contains the selection
         new_name = None
         obj_list = cmd.get_names('objects')
         for a in obj_list:
            if cmd.get_type(a)=="object:molecule":
               if cmd.count_atoms("(%s and %s)"%(a,sele_name)):
                  new_name = a
                  break
         src_frame = cmd.get_state()
         if new_name==None:
            print " Mutagenesis: object not found."
         else:
            auto_zoom = cmd.get_setting_text('auto_zoom')
            cmd.set('auto_zoom',"0",quiet=1)
            if self.mode!="current":
               # create copy w/o residue
               cmd.create(tmp_obj1,"(%s and not %s)"%(new_name,sele_name))
               # save copy for bonded atom reference
               cmd.create(tmp_obj3,new_name)
               # transfer the selection to copy
               cmd.select(sele_name,"(%s in %s)"%(tmp_obj3,sele_name))
               # create copy with mutant in correct frame
               cmd.create(tmp_obj2,obj_name,src_frame,1)
               cmd.delete(new_name)
               # create the merged molecule
               cmd.create(new_name,"(%s or %s)"%(tmp_obj1,tmp_obj2),1) # only one state in merged object...
               # now connect them
               cmd.bond("(%s in %s and n;N)"%(new_name,tmp_obj2),
                        "(name C and (%s in (neighbor %s)))"%
                        (new_name,sele_name))
               cmd.bond("(%s in %s and n;C)"%(new_name,tmp_obj2),
                        "(name N and (%s in (neighbor %s)))"%
                        (new_name,sele_name))
               # now transfer selection back to the modified object
               cmd.delete(tmp_obj1)
               cmd.delete(tmp_obj2)
               cmd.delete(tmp_obj3)
               self.clear()
               # and return to frame 1
               cmd.frame(1)
               cmd.refresh_wizard()               
            else:
               # create copy with conformation in correct state
               cmd.create(tmp_obj2,obj_name,src_frame,1)
               # save existing conformation on undo stack
#               cmd.edit("((%s in %s) and name ca)"%(new_name,sele_name))
               cmd.push_undo("("+sele_name+")")
               # modify the conformation
               cmd.update(new_name,tmp_obj2)
#               cmd.unpick()
               cmd.delete(tmp_obj2)
               self.clear()
               # and return to frame 1
               cmd.frame(1)
               cmd.refresh_wizard()                              
            cmd.set('auto_zoom',auto_zoom,quiet=1)
               
   def get_prompt(self):
      self.prompt = None
      if self.status==0:
         self.prompt = [ 'Pick a residue...']
      elif self.status==1:
         self.prompt = [ 'Select a conformational state, or pick a new residue...' ]
      return self.prompt

   def do_library(self):
      cmd.feedback("push")
      cmd.feedback("disable","selector","everythin")
      cmd.feedback("disable","editor","actions")

      self.prompt = [ 'Loading rotamers...']
      
      auto_zoom = cmd.get_setting_text('auto_zoom')
      cmd.set('auto_zoom',"0",quiet=1)
      cmd.frame(0)
      cmd.delete(tmp_name)
      if self.mode=="current":
         pymol.stored.resn=""
         cmd.iterate("(%s and n;ca)"%sele_name,"stored.resn=resn")
         res_type = pymol.stored.resn
         cmd.create(tmp_name,sele_name,1,1)         
      else:
         res_type = self.mode
         cmd.fragment(string.lower(self.mode),tmp_name)
         cmd.remove("("+tmp_name+" and hydro)")
         # copy identifying information
         cmd.iterate("(%s and n;ca)"%sele_name,"stored.chain=chain")
         cmd.alter("(%s)"%tmp_name,"chain=stored.chain")
         cmd.iterate("(%s and n;ca)"%sele_name,"stored.resi=resi")
         cmd.alter("(%s)"%tmp_name,"resi=stored.resi")
         cmd.iterate("(%s and n;ca)"%sele_name,"stored.segi=segi")
         cmd.alter("(%s)"%tmp_name,"segi=stored.segi")
         # move the fragment
         if ((cmd.count_atoms("(%s and n;cb)"%tmp_name)>0) and
             (cmd.count_atoms("(%s and n;cb)"%sele_name)>0)):
            cmd.pair_fit("(%s and n;ca)"%tmp_name,
                         "(%s and n;ca)"%sele_name,
                         "(%s and n;cb)"%tmp_name,
                         "(%s and n;cb)"%sele_name,
                         "(%s and n;c)"%tmp_name,
                         "(%s and n;c)"%sele_name,
                         "(%s and n;n)"%tmp_name,
                         "(%s and n;n)"%sele_name)
         else:
            cmd.pair_fit("(%s and n;ca)"%tmp_name,
                         "(%s and n;ca)"%sele_name,
                         "(%s and n;c)"%tmp_name,
                         "(%s and n;c)"%sele_name,
                         "(%s and n;n)"%tmp_name,
                         "(%s and n;n)"%sele_name)
         # fix the carbonyl position...
         cmd.iterate_state(1,"(%s and n;o)"%sele_name,"stored.list=[x,y,z]")
         cmd.alter_state(1,"(%s and n;o)"%tmp_name,"(x,y,z)=stored.list")
         
      cmd.delete(obj_name)
      if self.library.has_key(res_type):
         lib = self.library[res_type]
         state = 1
         for a in lib:
            cmd.create(obj_name,tmp_name,1,state)
            for b in a.keys():
               if b!='FREQ':
                  cmd.set_dihedral("(%s & n;%s)"%(obj_name,b[0]),
                                   "(%s & n;%s)"%(obj_name,b[1]),
                                   "(%s & n;%s)"%(obj_name,b[2]),
                                   "(%s & n;%s)"%(obj_name,b[3]),
                                   a[b],state=state)
               else:
                  cmd.set_title(obj_name,state,"%1.1f%%"%(a[b]*100))
            state = state + 1
         cmd.delete(tmp_name)
         print " Mutagenesis: %d conformations loaded."%len(self.library[res_type])
      else:
         cmd.create(obj_name,tmp_name,1,1)
         print " Mutagenesis: no additional conformations in library."
      pymol.util.cbaw(obj_name)
      cmd.hide("("+obj_name+")")
      cmd.show(self.rep,obj_name)
      cmd.show('lines',obj_name) # always show lines
      cmd.set('auto_zoom',auto_zoom,quiet=1)
      cmd.delete(tmp_name)
      cmd.frame(0)
      cmd.unpick()
      cmd.feedback("pop")
      
   def do_pick(self,bondFlag):
      if bondFlag:
         self.error = "Error: please select an atom, not a bond."
         print self.error
      else:
         if self.status!=0:
            cmd.delete(obj_name)
         cmd.select(sele_name,"(byres pk1)")
         cmd.unpick()
         cmd.enable(sele_name)
         self.status = 1
         self.error = None
         self.do_library()
      cmd.refresh_wizard()






#
#
#

import bond_amber
import protein_residues
import protein_amber

from chempy.neighbor import Neighbor
from chempy.models import Connected
from chempy import Bond,place,feedback
from chempy.cpv import *

MAX_BOND_LEN = 2.2
PEPT_CUTOFF = 1.7

N_TERMINAL_ATOMS = ('HT','HT1','HT2','HT3','H1','H2','H3',
                    '1H','2H','3H','1HT','2HT','3HT')

C_TERMINAL_ATOMS = ('OXT','O2','OT1','OT2')

#---------------------------------------------------------------------------------
# NOTE: right now, the only way to get N-terminal residues is to
# submit a structure which contains at least one N_TERMINAL hydrogens

def generate(model, forcefield = protein_amber, histidine = 'HIE',
             skip_sort=None, bondfield = bond_amber ):

   strip_atom_bonds(model) # remove bonds between non-hetatms (ATOM)
   add_bonds(model,forcefield=forcefield)   
   connected = model.convert_to_connected()
   add_hydrogens(connected,forcefield=forcefield,skip_sort=skip_sort)
   place.simple_unknowns(connected,bondfield = bondfield)
   return connected.convert_to_indexed()

#---------------------------------------------------------------------------------
def strip_atom_bonds(model):
   new_bond = []
   matom = model.atom
   for a in model.bond:
      if matom[a.index[0]].hetatm or matom[a.index[1]].hetatm:
         new_bond.append(a)
   model.bond = new_bond
   
#---------------------------------------------------------------------------------
def assign_types(model, forcefield = protein_amber, histidine = 'HIE' ):
   '''   
assigns types: takes HIS -> HID,HIE,HIP and CYS->CYX where appropriate
but does not add any bonds!
'''
   if feedback['actions']:
      print " "+str(__name__)+": assigning types..."
   if str(model.__class__) != 'chempy.models.Indexed':
      raise ValueError('model is not an "Indexed" model object')
   if model.nAtom:
      crd = model.get_coord_list()
      nbr = Neighbor(crd,MAX_BOND_LEN)
      res_list = model.get_residues()
      if len(res_list):
         for a in res_list:
            base = model.atom[a[0]]
            if not base.hetatm:
               resn = base.resn
               if resn == 'HIS':
                  for c in range(a[0],a[1]): # this residue
                     model.atom[c].resn = histidine
                  resn = histidine
               if resn == 'N-M': # N-methyl from Insight II, 
                  for c in range(a[0],a[1]): # this residue
                     model.atom[c].resn = 'NME'
                  resn = 'NME'
               # find out if this is n or c terminal residue
               names = []
               for b in range(a[0],a[1]):
                  names.append(model.atom[b].name)
               tmpl = protein_residues.normal
               if forcefield:
                  ffld = forcefield.normal
               for b in N_TERMINAL_ATOMS:
                  if b in names:
                     tmpl = protein_residues.n_terminal
                     if forcefield:
                        ffld = forcefield.n_terminal
                     break
               for b in C_TERMINAL_ATOMS:
                  if b in names:
                     tmpl = protein_residues.c_terminal
                     if forcefield:
                        ffld = forcefield.c_terminal
                     break
               if not tmpl.has_key(resn):
                  raise RuntimeError("unknown residue type '"+resn+"'")
               else:
                  # reassign atom names and build dictionary
                  dict = {}
                  aliases = tmpl[resn]['aliases']
                  for b in range(a[0],a[1]):
                     at = model.atom[b]
                     if aliases.has_key(at.name):
                        at.name = aliases[at.name]
                     dict[at.name] = b
                     if forcefield:
                        k = (resn,at.name)
                        if ffld.has_key(k):
                           at.text_type = ffld[k]['type']
                           at.partial_charge = ffld[k]['charge']
                        else:
                           raise RuntimeError("no parameters for '"+str(k)+"'")
                  if dict.has_key('SG'): # cysteine
                     cur = dict['SG']
                     at = model.atom[cur]
                     lst = nbr.get_neighbors(at.coord)
                     for b in lst:
                        if b>cur: # only do this once (only when b>cur - i.e. this is 1st CYS)
                           at2 = model.atom[b]
                           if at2.name=='SG':
                              if not at2.in_same_residue(at):
                                 dst = distance(at.coord,at2.coord)
                                 if dst<=MAX_BOND_LEN:
                                    if forcefield:
                                       for c in range(a[0],a[1]): # this residue
                                          atx = model.atom[c]
                                          atx.resn = 'CYX'
                                          resn = atx.resn
                                          if (c<=b):
                                             k = ('CYX',atx.name)
                                             if ffld.has_key(k):
                                                atx.text_type = ffld[k]['type']
                                                atx.partial_charge = ffld[k]['charge']
                                             else:
                                                raise RuntimeError("no parameters for '"+str(k)+"'")
                                       for d in res_list: # other residue
                                          if (b>=d[0]) and (b<d[1]):
                                             for c in range(d[0],d[1]):
                                                atx = model.atom[c]
                                                atx.resn = 'CYX'
                                                # since b>cur, assume assignment later on
                                 break
   
#---------------------------------------------------------------------------------
def add_bonds(model, forcefield = protein_amber, histidine = 'HIE' ):
   '''
add_bonds(model, forcefield = protein_amber, histidine = 'HIE' )

(1) fixes aliases, assigns types, makes HIS into HIE,HID, or HIP
    and changes cystine to CYX
(2) adds bonds between existing atoms
   '''
   if feedback['actions']:
      print " "+str(__name__)+": assigning types and bonds..."
   if str(model.__class__) != 'chempy.models.Indexed':
      raise ValueError('model is not an "Indexed" model object')
   if model.nAtom:
      crd = model.get_coord_list()
      nbr = Neighbor(crd,MAX_BOND_LEN)
      res_list = model.get_residues()
      if len(res_list):
         for a in res_list:
            base = model.atom[a[0]]
            if not base.hetatm:
               resn = base.resn
               if resn == 'HIS':
                  for c in range(a[0],a[1]): # this residue
                     model.atom[c].resn = histidine
                  resn = histidine
               if resn == 'N-M': # N-methyl from Insight II, 
                  for c in range(a[0],a[1]): # this residue
                     model.atom[c].resn = 'NME'
                  resn = 'NME'
               # find out if this is n or c terminal residue
               names = []
               for b in range(a[0],a[1]):
                  names.append(model.atom[b].name)
               tmpl = protein_residues.normal
               if forcefield:
                  ffld = forcefield.normal
               for b in N_TERMINAL_ATOMS:
                  if b in names:
                     tmpl = protein_residues.n_terminal
                     if forcefield:
                        ffld = forcefield.n_terminal
                     break
               for b in C_TERMINAL_ATOMS:
                  if b in names:
                     tmpl = protein_residues.c_terminal
                     if forcefield:
                        ffld = forcefield.c_terminal
                     break
               if not tmpl.has_key(resn):
                  raise RuntimeError("unknown residue type '"+resn+"'")
               else:
                  # reassign atom names and build dictionary
                  dict = {}
                  aliases = tmpl[resn]['aliases']
                  for b in range(a[0],a[1]):
                     at = model.atom[b]
                     if aliases.has_key(at.name):
                        at.name = aliases[at.name]
                     dict[at.name] = b
                     if forcefield:
                        k = (resn,at.name)
                        if ffld.has_key(k):
                           at.text_type = ffld[k]['type']
                           at.partial_charge = ffld[k]['charge']
                        else:
                           raise RuntimeError("no parameters for '"+str(k)+"'")
                  # now add bonds for atoms which are present
                  bonds = tmpl[resn]['bonds']
                  mbond = model.bond
                  for b in bonds.keys():
                     if dict.has_key(b[0]) and dict.has_key(b[1]):
                        bnd = Bond()
                        bnd.index = [ dict[b[0]], dict[b[1]] ]
                        bnd.order = bonds[b]['order']
                        mbond.append(bnd)
                  if dict.has_key('N'):  # connect residues N-C based on distance
                     cur_n = dict['N']
                     at = model.atom[cur_n]
                     lst = nbr.get_neighbors(at.coord)
                     for b in lst:
                        at2 = model.atom[b]
                        if at2.name=='C':
                           if not at2.in_same_residue(at):
                              dst = distance(at.coord,at2.coord)
                              if dst<=PEPT_CUTOFF:
                                 bnd=Bond()
                                 bnd.index = [cur_n,b]
                                 bnd.order = 1
                                 mbond.append(bnd)
                                 break
                  if dict.has_key('SG'): # cysteine
                     cur = dict['SG']
                     at = model.atom[cur]
                     lst = nbr.get_neighbors(at.coord)
                     for b in lst:
                        if b>cur: # only do this once (only when b>cur - i.e. this is 1st CYS)
                           at2 = model.atom[b]
                           if at2.name=='SG':
                              if not at2.in_same_residue(at):
                                 dst = distance(at.coord,at2.coord)
                                 if dst<=MAX_BOND_LEN:
                                    bnd=Bond()
                                    bnd.index = [cur,b]
                                    bnd.order = 1
                                    mbond.append(bnd)
                                    if forcefield:
                                       for c in range(a[0],a[1]): # this residue
                                          atx = model.atom[c]
                                          atx.resn = 'CYX'
                                          resn = atx.resn
                                          k = ('CYX',atx.name)
                                          if ffld.has_key(k):
                                             atx.text_type = ffld[k]['type']
                                             atx.partial_charge = ffld[k]['charge']
                                          else:
                                             raise RuntimeError("no parameters for '"+str(k)+"'")
                                       for d in res_list: 
                                          if (b>=d[0]) and (b<d[1]): # find other residue
                                             for c in range(d[0],d[1]):
                                                atx = model.atom[c]
                                                atx.resn = 'CYX'
                                                # since b>cur, assume assignment later on
                                    break
                     
#---------------------------------------------------------------------------------
def add_hydrogens(model,forcefield=protein_amber,skip_sort=None):
   # assumes no bonds between non-hetatms
   if feedback['actions']:
      print " "+str(__name__)+": adding hydrogens..."
   if str(model.__class__) != 'chempy.models.Connected':
      raise ValueError('model is not a "Connected" model object')
   if model.nAtom:
      if not model.index:
         model.update_index()
      res_list = model.get_residues()
      if len(res_list):
         for a in res_list:
            base = model.atom[a[0]]
            if not base.hetatm:
               resn = base.resn
               # find out if this is n or c terminal residue
               names = []
               for b in range(a[0],a[1]):
                  names.append(model.atom[b].name)
               tmpl = protein_residues.normal
               if forcefield:
                  ffld = forcefield.normal
               for b in N_TERMINAL_ATOMS:
                  if b in names:
                     tmpl = protein_residues.n_terminal
                     if forcefield:
                        ffld = forcefield.n_terminal
                     break
               for b in C_TERMINAL_ATOMS:
                  if b in names:
                     tmpl = protein_residues.c_terminal
                     if forcefield:
                        ffld = forcefield.c_terminal
                     break
               if not tmpl.has_key(resn):
                  raise RuntimeError("unknown residue type '"+resn+"'")
               else:
                  # build dictionary
                  dict = {}
                  for b in range(a[0],a[1]):
                     at = model.atom[b]
                     dict[at.name] = b
                  # find missing bonds with hydrogens
                  bonds = tmpl[resn]['bonds']
                  mbond = model.bond
                  for b in bonds.keys():
                     if dict.has_key(b[0]) and (not dict.has_key(b[1])):
                        at = model.atom[dict[b[0]]]
                        if at.symbol != 'H':
                           name = b[1]
                           symbol = tmpl[resn]['atoms'][name]['symbol']
                           if symbol == 'H':
                              newat = at.new_in_residue()
                              newat.name = name
                              newat.symbol = symbol
                              k = (resn,newat.name)
                              newat.text_type = ffld[k]['type']
                              newat.partial_charge = ffld[k]['charge']
                              idx1 = model.index[id(at)]
                              idx2 = model.add_atom(newat)
                              bnd = Bond()
                              bnd.index = [ idx1, idx2 ]
                              bnd.order = bonds[b]['order']
                              mbond[idx1].append(bnd)
                              mbond[idx2].append(bnd)
                     if (not dict.has_key(b[0])) and dict.has_key(b[1]):
                        at = model.atom[dict[b[1]]]
                        if at.symbol != 'H':
                           name = b[0]
                           symbol = tmpl[resn]['atoms'][name]['symbol']
                           if symbol == 'H':
                              newat = at.new_in_residue()
                              newat.name = name
                              newat.symbol = symbol
                              k = (resn,newat.name)
                              newat.text_type = ffld[k]['type']
                              newat.partial_charge = ffld[k]['charge']
                              idx1 = model.index[id(at)]
                              idx2 = model.add_atom(newat)
                              bnd = Bond()
                              bnd.index = [ idx1, idx2 ]
                              bnd.order = bonds[b]['order']
                              mbond[idx1].append(bnd)
                              mbond[idx2].append(bnd)
      if not skip_sort:
         model.sort()



# 
# basic run through the PDB with ray-tracing
#
      
from glob import glob

import threading
import time
import whrandom
from pymol import cmd
import sys, os, os.path


ent_dir = "pdb"


def load():
   r = 0
   rep = [ "lines","sticks","spheres","dots","ribbon","cartoon" ]
   list = glob("pdb/*/*")
   for file in list:
      print file
      cmd.delete('pdb')
      cmd.load(file,'pdb')
      cmd.orient('pdb')
      cmd.color('red','ss h')
      cmd.color('yellow','ss s')
      cmd.hide()
      cmd.show(rep[r],"all")
      r = r + 1
      if r>=len(rep): r=0;
      sys.__stderr__.write(".")
      sys.__stderr__.flush()
      n = cmd.count_states()
      cmd.ray(100,75)
      if n>1:
         sys.__stderr__.write("\n")
         sys.__stderr__.flush()
      cmd.refresh()
load()





#A* -------------------------------------------------------------------
#B* This file contains source code for the PyMOL computer program
#C* copyright 1998-2000 by Warren Lyford Delano of DeLano Scientific. 
#D* -------------------------------------------------------------------
#E* It is unlawful to modify or remove this copyright notice.
#F* -------------------------------------------------------------------
#G* Please see the accompanying LICENSE file for further information. 
#H* -------------------------------------------------------------------
#I* Additional authors of this source file include:
#-* Peter Haebel
#-* 
#-*
#Z* -------------------------------------------------------------------

import os
import traceback

povray_exe = "x-povray"

def render_from_string(header,pov_inp,prefix,width,height,antialias):
   r = None
   try:
      pov = prefix +".pov"
      png = prefix +".png"
      f=open(pov,'w')
      f.write(header)
      f.write(pov_inp)
      f.close()
      if os.path.exists(png):
         os.unlink(png)
      if antialias:
         ant="+A"
      else:
         ant=""
      os.system("%s +I%s +O%s +W%d +H%d %s"%(
                povray_exe,pov,png,width,height,ant))
      if os.path.exists(png):
         r = 1
   except:
      traceback.print_exc()
   return r
